
from .. import geometry
from .. import algebra
import numpy as np

def generate_dummy_hamiltonian(d):
    """Create a dummy Hamiltonian from a dictionary"""
    g = geometry.chain() # create a chain
    h = g.get_hamiltonian() # create a dummy Hamiltonian
    h = h.get_multicell() # set as multicell Hamiltonian
    from ..multihopping import MultiHopping
    if type(d) is not dict: raise
    d = clean_dict(d) # overwrite
    h.set_multihopping(MultiHopping(d)) # set the dictionary
    h.geometry.supercell(h.intra.shape[0]) # dimensionality
    dim = get_dimensionality(d)
    h.geometry.dimensionality = dim
    h.geometry.a1 = np.array([1.,0.,0.])
    h.geometry.a2 = np.array([0.,1.,0.])
    h.geometry.a3 = np.array([0.,0.,1.])
    h.dimensionality = dim
    return h


def clean_dict(d):
    """Given a dictionary, return a cleaned up dictionary"""
    dout = dict()
    for key in d:
        key2 = tuple([int(k) for k in np.array(key)])
        dout[key2] = algebra.todense(np.array(d[key])) # to dense
    return dout


def get_dimensionality(d):
    """This function will try to guess the dimensionality of the system"""
    dim = 0 # zero dimensional
    vs = np.array([np.array(key) for key in d])
    from numpy.linalg import matrix_rank
    return matrix_rank(vs)
    





def modify_hamiltonian_matrices(self,f0,use_geometry=False):
    """Apply a certain function to all the matrices"""
    # wrapper function
    if use_geometry: # use geometry
        g = self.geometry # get the geometry
        def f(m,dr): 
            return f0(m,g.r,g.replicas(d=dr)) # use r1 and r2
    else:
        def f(m,dr): return f0(m) # do not use geometry
    # apply the function
    self.intra = f(self.intra,[0,0,0]) # modify intracell
    if self.is_multicell: # for multicell hamiltonians
      for i in range(len(self.hopping)): # loop over hoppings
        # modify Hamiltonian matrix
        self.hopping[i].m = f(self.hopping[i].m,self.hopping[i].dir) 
    else: # conventional way
      if self.dimensionality==0: pass # one dimensional systems
      elif self.dimensionality==1: # one dimensional systems
        self.inter = f(self.inter,[1,0,0])
      elif self.dimensionality==2: # two dimensional systems
        self.tx = f(self.tx,[1,0,0])
        self.ty = f(self.ty,[0,1,0])
        self.txy = f(self.txy,[1,1,0])
        self.txmy = f(self.txmy,[1,-1,0])
      else: raise




###### functions to generate julia inputs #####


def write_hamiltonian(h):
  """Writes a hamiltonian in julia_hamiltonian.jl as a julia script"""
  def spm(m):
    """Non vanishing elements"""
    from scipy.sparse import coo_matrix as coo
    mc = coo(m) # to coo_matrixi
    data = mc.data # get data
    col = mc.col # get column index
    row = mc.row # get row index
    return (row,col,data.real,data.imag) # return things to print

  if h.dimensionality==1:
    f = open("julia_hamiltonian.jl","w") # open file
    n = len(h.intra) # length of the hamiltonian
    f.write("module julia_hamiltonian\n# dimension\n")  
    f.write("using hamiltonians\n")  
    f.write("function get_hamiltonian()\n")  
    f.write("  h =hamiltonians.hamiltonian1d()\n")  
    f.write("  intra = im*zeros("+str(n)+","+str(n)+")\n") # create intra term
    f.write("  inter = 0.*intra\n\n") # create inter term
    f.write("  # elements of the hamiltonian\n") 
    def write_matrix(m,name):
      """Writes aparticular matrix"""
      row,col,re,im = spm(m)
      for (i,j,a,b) in zip(row,col,re,im):
        f.write("  "+name+"["+str(i+1)+","+str(j+1)+"] = "+
                        str(a)+" + im*"+str(b)+"\n")
      f.write("  ###############\n\n")
    write_matrix(h.intra,"intra")
    write_matrix(h.inter,"inter")

    f.write("  h.intra = intra\n  h.inter = inter\n")  
    f.write("  return h\n")  
    f.write("end\n")  
    f.write("end\n")  
    f.close()  # close file





from __future__ import print_function

def effective_hamiltonian(h,k=[0.,0.,0.],ewindow=None):
  """Return an effective k dot p Hamiltonian at kpoint k"""
  if h.dimensionality != 2: raise # only for 2d




def fermi_velocity(h,n,k=[0.,0.,0.]):
    """Return the Fermi velocity at the Fermi energy"""
    from current import current_operator






from __future__ import print_function
import numpy as np

def bulk2ribbon(g, n=5):
  """ Transformas a 2D geometry into a ribbon"""
  if not g.dimensionality == 2: raise # has to be two dimensional
  if not np.abs(g.a1.dot(g.a2)) < 0.0001: raise # has to be orthogonal
  go = g.copy() # create new geometry
  go.dimensionality = 1 # ribbon
  go.x = []
  go.y = []
  go.z = []
  for i in range(n):
    go.x += (g.x).tolist() # append x
    go.z += (g.z).tolist() # append x
    go.y += (g.y+i*g.a2[1]).tolist() # append x
  go.x = np.array(go.x)
  go.y = np.array(go.y)
  go.z = np.array(go.z)
  go.xyz2r()
  go.celldis = g.a1[0]
  go.center()
  return go


def reflect(g):
  """ Reflects a certain geometry with respect to the origin"""
  g.y = g.y - min(g.y) # move to the zero
  go = g.copy()
  go.x = (g.x).tolist() + (g.x).tolist()
  go.y = (g.y).tolist() + (-g.y).tolist()
  go.z = (g.z).tolist() + (g.z).tolist()
  go.x = np.array(go.x)
  go.y = np.array(go.y)
  go.z = np.array(go.z)
  go.xyz2r()
  go = remove_repeated(go) # remove the repeted atoms
  return go  




def remove_repeated(g):
  """ Remove repeated coordinates"""
  go = g.copy() # copy geometry
  go.r = [] # empty list
  for i in range(len(g.r)): # only accept
    unique = True
    for j in range(len(go.r)):
      if i!=j:
        dr = g.r[i] -g.r[j]
        dr = dr.dot(dr) # distance
        if dr<0.2:
          unique=False
    if unique:
      go.r.append(g.r[i])
    else:
      print("Repeated",i)
  go.r = np.array(go.r)
  go.r2xyz()
  return go


 

def hamiltonian_bulk2ribbon(h,n=10,sparse=False,check=True):
  """Converts a hexagonal bulk hamiltonian into a ribbon hamiltonian"""
  from scipy.sparse import csc_matrix as csc
  from scipy.sparse import bmat
  go = h.geometry.copy() # copy geometry
  ho = h.copy() # copy hamiltonian
#  if np.abs(go.a1.dot(go.a2))>0.01: 
#    if check: raise # if axis non orthogonal
  ho.dimensionality = 1 # reduce dimensionality
  ho.geometry.dimensionality = 1 # reduce dimensionality
  intra = [[None for i in range(n)] for j in range(n)]
  inter = [[None for i in range(n)] for j in range(n)]
  for i in range(n): # to the the sam index
    intra[i][i] = csc(h.intra)
    inter[i][i] = csc(h.tx)
  for i in range(n-1): # one more or less
    intra[i][i+1] = csc(h.ty)
    intra[i+1][i] = csc(h.ty.H)
    inter[i+1][i] = csc(h.txmy)
    inter[i][i+1] = csc(h.txy)
  if sparse:
    ho.intra = bmat(intra)
    ho.inter = bmat(inter)
    ho.is_sparse = True # hamiltonian is sparse
  else:
    ho.intra = bmat(intra).todense()
    ho.inter = bmat(inter).todense()
  # calculate the angle 
  import sculpt
  ho.geometry = sculpt.rotate_a2b(ho.geometry,ho.geometry.a1,np.array([1.,0.,0.]))
  ho.geometry = h.geometry.supercell((1,n)) # create supercell
  ho.geometry.dimensionality = 1
  ho.geometry.a1 = h.geometry.a1 # add the unit cell vector
  ho.dimensionality = 1
  # for geometries with names
  if ho.geometry.atoms_have_names:
    ho.geometry.atoms_names = ho.geometry.atoms_names*n
  return ho





from __future__ import print_function, division
import numpy as np
from numba import jit
from . import parallel

try:
  from . import density_matrixf90
  use_fortran = True
except:
  use_fortran = False
#  print("Fortran routines not working in densitymatrix.py")
 
use_fortran = False

def full_dm(h,use_fortran=True,nk=10,fermi=0.0,delta=1e-2,ds=None):
  if h.dimensionality == 0: fac = 1.
  elif h.dimensionality == 1: fac = 1./nk
  elif h.dimensionality == 2: fac = 1./nk**2
  elif h.dimensionality == 3: fac = 1./nk**3
  else: raise
  if ds is None: # no directions required
    es,vs = h.get_eigenvectors(nk=nk) # get eigenvectors
    es = es - fermi # shift by the Fermi energy
    if use_fortran:
      dm = density_matrixf90.density_matrix(np.array(es),np.array(vs),delta)
      return dm*fac
    else:
      return np.matrix(full_dm_python(h.intra.shape[0],es,np.array(vs)))*fac # call hte function
  else: # directions required
    es,vs,ks = h.get_eigenvectors(nk=nk,kpoints=True) # get eigenvectors
    es = es - fermi # shift by the Fermi energy
    ks = np.array(ks) # to array
    n = h.intra.shape[0] # dimensionality
    out = parallel.pcall(lambda x: full_dm_python_d(n,es,vs,ks,x)*fac,ds)
#    out = [full_dm_python_d(n,es,vs,ks,d)*fac for d in ds] # compute all the DM
    outd = dict() # dictionary
    for i in range(len(ds)): outd[tuple(ds[i])] = out[i] # as dictionary
    return outd
#    return out # return all the density matrices



def full_dm_python(n,es,vs):
  """Calculate the density matrix"""
  dm = np.zeros((n,n),dtype=np.complex)
  return full_dm_python_jit(n,es,vs,dm)


def full_dm_python_d(n,es,vs,ks,d):
  """Calculate the density matrix"""
  dm = np.zeros((n,n),dtype=np.complex)
  return full_dm_python_d_jit(n,es,vs,ks,np.array(d),dm)



@jit(nopython=True)
def full_dm_python_jit(n,es,vs,dm):
  """Auxiliary function to compute the density matrix"""
  for ie in range(len(es)): # loop
    if es[ie]<0.: # if below Fermi energy
      for i in range(n):
        for j in range(n): 
          dm[i,j] = dm[i,j] + vs[ie][i].conjugate()*vs[ie][j] # add contribution
  return dm


@jit(nopython=True)
def full_dm_python_d_jit(n,es,vs,ks,d,dm):
  """Auxiliary function to compute the density matrix"""
  for ie in range(len(es)): # loop
    k = ks[ie] # get kpoint
    kd = k[0]*d[0] + k[1]*d[1] + k[2]*d[2] # compute scalar product
    phi = np.exp(1j*np.pi*kd*2) # compute phase
#    phi = 0.0
    if es[ie]<0.: # if below Fermi energy
      for i in range(n):
        for j in range(n):
          dm[i,j] = dm[i,j] + phi*vs[ie][i].conjugate()*vs[ie][j] # add contribution
  return dm













def restricted_dm(h,use_fortran=True,mode="KPM",pairs=[],
                   scale=10.0,npol=400,ne=None):
  """Calculate certain elements of the density matrix"""
  if h.dimensionality != 0 : raise
  if mode=="full": # full inversion and then select
    dm = full_dm(h,use_fortran=use_fortran) # Full DM
    outm = np.array([dm[j,i] for (i,j) in pairs]) # get the desired ones
    return outm # return elements
  elif mode=="KPM": # use Kernel polynomial method
    if ne is None: ne = npol*4
    from . import kpm
    xin = np.linspace(-.99*scale,0.0,ne) # input x array
    out = np.zeros(len(pairs),dtype=np.complex)
    ii = 0
    for (i,j) in pairs: # loop over inputs
      (x,y) = kpm.dm_ij_energy(h.intra,i=i,j=j,scale=scale,npol=npol,
                      ne=ne,x=xin)
      out[ii] = np.trapz(y,x=x)/np.pi # pi is here so it normalizes to 0.5
      ii += 1
    return out
  else: raise
       
from . import algebra

def occupied_projector(m,delta=0.0):
    """Return a projector onto the occupied states"""
    (es,vs) = algebra.eigh(m) # diagonalize
    vs = vs.T # transpose
#    vs = vs[es<0.0] # occupied states
    if use_fortran:
      dm = density_matrixf90.density_matrix(np.array(es),np.array(vs),delta)
      return np.array(dm)
    else:
      return np.array(full_dm_python(m.shape[0],es,np.array(vs)))




import numpy as np
from scipy.integrate import dblquad 

def keldysh(v0=1.0,alpha=1.0,rcut=10.):
    """Keldysh potential from Phys. Rev. B 84, 085406"""

    def vf(r):
        r0 = alpha
        out = np.log(r/(r+r0)) + (0.5772 - np.log(2))*np.exp(-r/r0)
        return -v0*out/r0
    rs = np.linspace(1.,rcut,100)
    vr = [vf(ir) for ir in rs]
    
    from scipy.interpolate import interp1d
    
    fint = interp1d(rs,vr,fill_value=0.0,bounds_error=False)
#    return fint
    def fout(r1,r2=np.array([0.,0.,0.])):
        dr = r1-r2
        dr2 = dr.dot(dr)
        return fint(np.sqrt(dr2))
    return fout # return function




from ..geometry import same_site
import numpy as np
from ..kanemele import get_haldane_function
from ..utilities import get_callable
from .dvector import dvector2delta


def pairing_generator(self,delta=0.0,mode="swave",d=[0.,0.,1.],
    **kwargs):
    """Create a generator, taking as input two positions, and returning
    the 2x2 pairing matrix"""
    # wrapper for the amplitude and d-vector
    deltaf = get_callable(delta) # callable for the amplitude
    df = get_callable(d) # callable for the d-vector
    if callable(mode):
        weightf = mode # mode is a function returning a 2x2 pairing matrix
    elif mode=="swave":
        weightf = lambda r1,r2: same_site(r1,r2)*np.identity(2)
    elif mode=="triplet": 
        weightf = lambda r1,r2: pwave(r1,r2,df,**kwargs)
    elif mode=="nodal_fwave":
        weightf = lambda r1,r2: nodal_fwave(r1,r2,df,**kwargs)
    elif mode=="antihaldane":
        f = get_haldane_function(self.geometry,stagger=True)
        weightf = lambda r1,r2: f(r1,r2)*np.identity(2)
    elif mode=="haldane":
        f = get_haldane_function(self.geometry,stagger=False)
        weightf = lambda r1,r2: f(r1,r2)*np.identity(2)
    elif mode=="swavez":
        weightf = lambda r1,r2: same_site(r1,r2)*tauz
    elif mode=="px":
        weightf = lambda r1,r2: px(r1,r2)
    elif mode=="dpid":
        weightf = lambda r1,r2: dpid(r1,r2,**kwargs)
    elif mode=="swaveA":
        weightf = lambda r1,r2: swaveA(self.geometry,r1,r2)
    elif mode=="swaveB":
        weightf = lambda r1,r2: swaveB(self.geometry,r1,r2)
    elif mode=="swavesublattice":
        def weightf(r1,r2):
          return swaveB(self.geometry,r1,r2) - swaveA(self.geometry,r1,r2)
    elif mode=="dx2y2":
        weightf = lambda r1,r2: dx2y2(r1,r2)
    elif mode=="dxy":
        weightf = lambda r1,r2: dxy(r1,r2)
    elif mode=="snn":
        weightf = lambda r1,r2: swavenn(r1,r2)
    elif mode=="C3nn":
        weightf = lambda r1,r2: C3nn(r1,r2)
    elif mode=="SnnAB":
        weightf = lambda r1,r2: SnnAB(self.geometry,r1,r2)
    else: raise
    matrixf = lambda r1,r2: deltaf((r1+r2)/2.)*weightf(r1,r2) 
    return matrixf # return function

# matrices for the e-h subsector
iden = np.array([[1.,0.],[0.,1.]],dtype=np.complex)
tauz = np.array([[1.,0.],[0.,-1.]],dtype=np.complex)
taux = np.array([[0.,1.],[1.,0.]],dtype=np.complex)
tauy = np.array([[0.,1j],[-1j,0.]],dtype=np.complex)
UU = taux + 1j*tauy # projector in the UU sector
DD = taux - 1j*tauy # projector in DD sector


def swavenn(r1,r2):
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        return 1.0*iden
    return 0.0*iden


def dx2y2(r1,r2):
    """Function with first neighbor dx2y2 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        return (dr[0]**2 - dr[1]**2)*iden
    return 0.0*iden


def dxy(r1,r2):
    """Function with first neighbor dx2y2 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        return (dr[0]*dr[1])*iden
    return 0.0*iden





def C3nn(r1,r2):
    """Function with first neighbor C3 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
#        return dr[0]
        phi = np.arctan2(dr[1],dr[0]) # angle
        return 1.0*np.exp(1j*phi)*tauz
    return 0.0*tauz



def C3nn(r1,r2):
    """Function with first neighbor C3 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
#        return dr[0]
        phi = np.arctan2(dr[1],dr[0]) # angle
        return 1.0*np.exp(1j*phi)*tauz
    return 0.0*tauz



def SnnAB(g,r1,r2):
    """Swave between AB"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        i = g.get_index(r1,replicas=True)
        j = g.get_index(r2,replicas=True)
        if g.sublattice[i]==1 and g.sublattice[j]==-1:
          return 1.0*tauz
        else: return -1.0*tauz
#          return 1.0*np.matrix([[1.,0.],[0.,0.]])
    return 0.0*iden




def swaveA(g,r1,r2):
    """Swave only in A"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if dr2<0.001: # first neighbor
        i = g.get_index(r1,replicas=False)
        if i is None: return 0.0
        if g.sublattice[i]==1:
          return 1.0*iden
    return 0.0*iden


def swaveB(g,r1,r2):
    """Swave only in A"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if dr2<0.001: # first neighbor
        i = g.get_index(r1,replicas=False)
        if i is None: return 0.0
        if g.sublattice[i]==-1:
          return 1.0*iden
    return 0.0*iden


def px(r1,r2):
    """Function with first neighbor px profile"""
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: return dr[0]*tauz
    return 0.0*tauz


def get_triplet(r1,r2,df,L=1):
    """Function for p-wave order"""
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if abs(L)%2!=1: raise
    if 0.99<dr2<1.001: 
        phi = np.arctan2(dr[1],dr[0])
        d = df((r1+r2)/2.) # evaluate dvector
        delta = dvector2delta(d) # compute the local deltas
        ms = np.array([[delta[2],delta[0]],[delta[1],-delta[2]]])
        return np.exp(1j*phi*L)*ms
    else: return 0.0*tauz


def get_singlet(r1,r2,L=2):
    """Function for p-wave order"""
    if L%2!=0: raise
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if 0.99<dr2<1.001:
        phi = np.arctan2(dr[1],dr[0])
        out = np.exp(1j*phi*L)
#        print(out)
        return out*iden
    else: return 0.0*tauz


def pwave(*args,**kwargs): 
    return get_triplet(*args,L=1,**kwargs)


def dpid(*args,**kwargs):
    return get_singlet(*args,L=2,**kwargs)


def nodal_fwave(*args,**kwargs): 
    return get_triplet(*args,L=3,**kwargs) + get_triplet(*args,L=-3,**kwargs)




# special band structures

from __future__ import print_function
import scipy.linalg as lg
from scipy.sparse import csc_matrix
import scipy.sparse.linalg as slg
from scipy.sparse import issparse
import numpy as np
from . import timing
from . import klist
from . import topology
from . import operators
from .algebra import braket_wAw
from . import algebra

from .limits import densedimension as maxdim
arpack_tol = 1e-8
arpack_maxiter = 10000


def berry_bands(h,klist=None,mode=None,operator=None):
  """Calculate band structure resolved with Berry curvature"""
  ks = [] # list of kpoints
  if mode is not None: # get the mode
    if mode=="sz": operator = operators.get_sz(h)
    else: raise

  fo = open("BANDS.OUT","w")
  for ik in range(len(klist)): # loop over kpoints
    (es,bs) = topology.operator_berry_bands(h,k=klist[ik],operator=operator)
    for (e,b) in zip(es,bs):
      fo.write(str(ik)+"    "+str(e)+"    "+str(b)+"\n")
  fo.close()



def current_bands(h,klist=None):
  """Calcualte the band structure, with the bands"""
  if h.dimensionality != 1: raise # only 1 dimensional
  # go for the rest
  hkgen = h.get_hk_gen() # get generator of the hamiltonian
  if klist is None:  klist = np.linspace(0,1.,100) # generate k points
  fo = open("BANDS.OUT","w") # output file
  from . import current
  fj = current.current_operator(h) # function that generates the operator
  for k in klist: # loop over kpoints
    hk = hkgen(k) # get k-hamiltonian
    jk = fj(k) # get current operator
    evals,evecs = algebra.eigh(hk) # eigenvectors and eigenvalues
    evecs = np.transpose(evecs) # transpose eigenvectors
    for (e,w) in zip(evals,evecs): # do the loop
        waw = braket_wAw(w,jk).real # product
        fo.write(str(k)+"    "+str(e)+"   "+str(waw)+"\n")
  fo.close()









def ket_Aw(A,w):
  return A@w





def get_bands_nd(h,kpath=None,operator=None,num_bands=None,
                    callback=None,central_energy=0.0,nk=400,
                    output_file="BANDS.OUT",write=True,
                    silent=True):
  """
  Get an n-dimensional bandstructure
  """
  if num_bands is not None:
    if num_bands>(h.intra.shape[0]-1): num_bands=None
  if type(operator)==str: operator = h.get_operator(operator)
  if num_bands is None: # all the bands
    if operator is not None: 
      def diagf(m): # diagonalization routine
          return algebra.eigh(m) # all eigenvals and eigenfuncs
    else: 
      def diagf(m): # diagonalization routine
          return algebra.eigvalsh(m) # all eigenvals and eigenfuncs
  else: # using arpack
    h = h.copy()
    h.turn_sparse() # sparse Hamiltonian
    def diagf(m):
      eig,eigvec = slg.eigsh(m,k=num_bands,which="LM",sigma=central_energy,
                                  tol=arpack_tol,maxiter=arpack_maxiter)
      if operator is None: return eig
      else: return (eig,eigvec)
  # open file and get generator
  hkgen = h.get_hk_gen() # generator hamiltonian
  kpath = h.geometry.get_kpath(kpath,nk=nk) # generate kpath
  def getek(k):
    """Compute this k-point"""
    out = "" # output string
    hk = hkgen(kpath[k]) # get hamiltonian
    if operator is None:
      es = diagf(hk)
      es = np.sort(es) # sort energies
      for e in es:  # loop over energies
        out += str(k)+"   "+str(e)+"\n" # write in file
      if callback is not None: callback(k,es) # call the function
    else:
      es,ws = diagf(hk)
      ws = ws.transpose() # transpose eigenvectors
      def evaluate(w,k,A): # evaluate the operator
          if type(A)==operators.Operator:
              waw = A.braket(w,k=kpath[k]).real
          elif callable(A):  
            try: waw = A(w,k=kpath[k]) # call the operator
            except: 
              print("Check out the k optional argument in operator")
              waw = A(w) # call the operator
          else: waw = braket_wAw(w,A).real # calculate expectation value
          return waw # return the result
      for (e,w) in zip(es,ws):  # loop over waves
        if isinstance(operator, (list,)): # input is a list
            waws = [evaluate(w,k,A) for A in operator]
        else: waws = [evaluate(w,k,operator)]
        out += str(k)+"   "+str(e)+"  " # write in file
        for waw in waws:  out += str(waw)+"  " # write in file
        out += "\n" # next line
      # callback function in each iteration
      if callback is not None: callback(k,es,ws) # call the function
    return out # return string
  ### Now evaluate the function
  from . import parallel
  if write: f = open(output_file,"w") # open bands file
  if parallel.cores==1: ### single thread ###
    tr = timing.Testimator("BANDSTRUCTURE",silent=silent) # generate object
    esk = "" # empty list
    for k in range(len(kpath)): # loop over kpoints
      tr.remaining(k,len(kpath)) # estimate of the time
      ek = getek(k)
      esk += ek # store
      if write: f.write(ek) # write this kpoint
      if write: f.flush() # flush in file
  else: # parallel run
      esk = parallel.pcall(getek,range(len(kpath))) # compute all
      esk = "".join(esk) # concatenate all
      if write: f.write(esk)
  if write: f.close()
#  print("\nBANDS finished")
  esk = esk.split("\n") # split
  del esk[-1] # remove last one
  esk = np.array([[float(i) for i in ek.split()] for ek in esk]).T
  return esk # return data



def smalleig(m,numw=10,evecs=False):
  """
  Return the smallest eigenvalues using arpack
  """
  tol = arpack_tol
  eig,eigvec = slg.eigsh(m,k=numw,which="LM",sigma=0.0,
                                  tol=tol,maxiter=arpack_maxiter)
  if evecs:  return eig,eigvec.transpose()  # return eigenvectors
  else:  return eig  # return eigenvalues


def lowest_bands(h,nkpoints=100,nbands=10,operator = None,
                   info = False,kpath = None,discard=None):
  """
  Returns the lowest eigenvaleus of the system
  """
  from scipy.sparse import csc_matrix
  if kpath is None: 
    k = klist.default(h.geometry) # default path
  else: k = kpath
  import gc # garbage collector
  fo = open("BANDS.OUT","w")
  if operator is None: # if there is not an operator
    if h.dimensionality==0:  # dot
      eig,eigvec = slg.eigsh(csc_matrix(h.intra),k=nbands,which="LM",sigma=0.0,
                                  tol=arpack_tol,maxiter=arpack_maxiter)
      eigvec = eigvec.transpose() # transpose
      iw = 0
      for i in range(len(eig)):
        if discard is not None: # use the function
          v = eigvec[i] # eigenfunction
          if discard(v): continue
        fo.write(str(iw)+"     "+str(eig[i])+"\n")
        iw += 1 # increase counter
    elif h.dimensionality>0: 
      hkgen = h.get_hk_gen() # get generator
      for ik in k:  # ribbon
        hk = hkgen(ik) # get hamiltonians
        gc.collect() # clean memory
        eig,eigvec = slg.eigsh(hk,k=nbands,which="LM",sigma=0.0)
        del eigvec # clean eigenvectors
        del hk # clean hk
        for e in eig:
          fo.write(str(ik)+"     "+str(e)+"\n")
        if info:  print("Done",ik,end="\r")
    else: raise # ups
  else:  # if there is an operator
    if h.dimensionality==1:
      hkgen = h.get_hk_gen() # get generator
      for ik in k:
        hk = hkgen(ik) # get hamiltonians
        eig,eigvec = slg.eigsh(hk,k=nbands,which="LM",sigma=0.0)
        eigvec = eigvec.transpose() # tranpose the matrix
        if info:  print("Done",ik,end="\r")
        for (e,v) in zip(eig,eigvec): # loop over eigenvectors
          a = braket_wAw(v,operator)
          fo.write(str(ik)+"     "+str(e)+"     "+str(a)+"\n")
  fo.close()






import numpy as np
from scipy.sparse import csc_matrix

def add_sublattice_imbalance(h,mass):
    """ Adds to the intracell matrix a sublattice imbalance """
    if h.geometry.has_sublattice:  # if has sublattice
      def ab(i):
        return h.geometry.sublattice[i]
    else:
      print("WARNING, no sublattice present")
      return 0. # if does not have sublattice
    natoms = len(h.geometry.r) # assume spinpolarized calculation 
    rows = range(natoms)
    if callable(mass):  # if mass is a function
      r = h.geometry.r
      data = [mass(r[i])*ab(i) for i in range(natoms)]
    else: data = [mass*ab(i) for i in range(natoms)]
    massterm = csc_matrix((data,(rows,rows)),shape=(natoms,natoms)) # matrix
    h.intra = h.intra + h.spinless2full(massterm)






import numpy as np
from . import parallel
from . import algebra
from . import magnetism

def dominant_correlation(h0,filling=0.5,dm=1e-1,
        write=False,**kwargs):
    """Compute the dominant magnetic correlator"""
    h = h0.copy() # copy hamiltonian
    h.turn_dense()
    h.set_filling(filling) # set the desired filling
    if not h.has_spin: raise # only for spinful
    n = len(h.geometry.r) # number of sites
    def getrow(ii): # compute a row of the susceptibility matrix
        hi = h.copy() # copy Hamiltonian
        ms = [[0.,0.,0.] for j in range(n)]
        ms[ii] = [0.,0.,dm] # perturbation
        magnetism.add_magnetism(hi,ms) # add local exchange fields
        hi.set_filling(filling,**kwargs) # set the desired filling
        print("Done correlator site",ii)
        out = hi.compute_vev("sz",delta=dm/2.,**kwargs) # compute magnetization
        return -np.array(out)/dm # return output
    out = parallel.pcall(getrow,range(n)) # perform all the computations
    from . import rkky
#    out = rkky.rkky0d(h)
    out = np.array(out) # as array
#    print("Next")
    print(np.max(np.abs(out-out.T)),np.mean(np.abs(out)))
    out = (out + out.T)/2.0 # make Hermitian
    (es,vs) = algebra.eigh(out) # diagonalize
    print(es)
    chi = vs.T[vs.shape[0]-1] # biggest eigenvector
    if write: h.geometry.write_profile(chi,name="CHI_PROFILE.OUT")
    return chi








#def magnetic_chi_ij_perturation_theory(h0,filling=0.5,dm=1e-1,write=False)



import scipy.linalg as lg


## routines for diagonalization ##

error = 1e-7

def eigh(m):
    if np.max(np.abs(m.imag))<error: # assume real
        print("Real")
        return lg.eigh(m.real) # diagonalize real matrix
    else: return lg.eigh(m) # diagonalize complex matrix






from __future__ import print_function
import numpy as np
import scipy.linalg as lg



def degenerate_points(h,n=0):
    """Return the points in the Brillouin zone that have a node
    in the bandstructure"""
    from scipy.optimize import differential_evolution
    bounds = [(0.,1.) for i in range(h.dimensionality)]
    hk_gen = h.get_hk_gen() # generator
    def get_point(x0):
      def f(k): # conduction band eigenvalues
        hk = hk_gen(k) # Hamiltonian
        es = lg.eigvalsh(hk) # get eigenvalues
        return abs(es[n]-es[n+1]) # gap
      res = differential_evolution(f,bounds=bounds) # minimize
      return res.x
    x0 = np.random.random(h.dimensionality) # inital vector
    return get_point(x0) # get the k-point


def dirac_points(h,n=0,dk=0.01):
    """Look for Dirac points in a Hamiltonian"""
    k = degenerate_points(h,n=n)
    from .topology import berry_phase
    kpath = [[np.cos(i), np.sin(i)] for i in np.pi*np.linspace(0.,2.,10)]
    kpath = [k+dk*np.array(ik) for ik in kpath] # loop
    b = berry_phase(h,kpath=kpath)/np.pi # normalize
    print(k,b)





# routines for dealing with spin models, using the Holfstein Primakov
# transformation

from . import neighbor
import numpy as np
from scipy.sparse import csc_matrix,coo_matrix
from scipy.sparse import identity as sparseiden

class SpinModel():
  def __init__(self,g):
    self.geometry = g # store the geometry
    self.dimensionality = g.dimensionality
    self.spins = np.array([1. for r in g.r]) # ones
    self.magnetization = np.array([[0.,0.,1.] for r in g.r]) # ones
  def hp_heisenberg(self,fun=None,d=None,k=None):
    return hp_heisenberg(self,fun=fun,d=d,k=k)
  def szsz(self,fun=None):
    return szsz(self,fun=fun)
  def setup_exchange(self):
    h = self.get_hamiltonian(has_spin=False) # generate a hamiltonian
    class Exchange(): pass
    self.exchange = Exchange() 
    self.exchange.intra = h.intra
    if self.dimensionality==0: pass
    elif self.dimensionality==1:
      self.exchange.inter = h.inter
    elif self.dimensionality==2:
      self.exchange.tx = h.tx
      self.exchange.ty = h.ty
      self.exchange.txy = h.txy
      self.exchange.txmy = h.txmy
    else: raise

 

def set_couplings(g,f=None):
  """Set the exchange couplings"""
  js = []
  xs = []
  ys = []
  if callable(f): # callable function
    m = neighbor.parametric_hopping(g.r,g.r)
  else: raise
  return m # return pairs

def xyj(m):
  from scipy.sparse import coo_matrix
  m2 = coo_matrix(m)
  return (m.rows,m.cols,m.data)


def hp_heisenberg(sm,fun=None,d=None,k=None):
  """Perform the Holfstein-Primakov for this spin model"""
  sm.hamiltonian = sm.geometry.get_hamiltonian(has_spin=False,fun=fun) # initialize
  spins = sm.spins # spins
  ns = len(spins) # number of sites
  inds = [i for i in range(ns)]
  def c2h(m0):
    def genij(i,j): return np.identity(3)*m0[i,j]
    return sites2coupling(genij,spins)
  # intra term
  for i in range(sm.hamiltonian.intra.shape[0]): 
    sm.hamiltonian.intra[i,i] = 0.0 # set coupling to zero
  (mons,mhop) = c2h(sm.hamiltonian.intra)
  if d is not None: # add the uniaxial anisotropy
    try: # for an array
      d[0]
      ds = d
    except: # for a number
      ds = [d for i in range(ns)] # loop over spins
    mani =  csc_matrix((ds,(inds,inds)),shape=(ns,ns)).todense()
    mons = mons + mani # add anisotropy 
  sm.hamiltonian.intra = mons + mhop
  # now add the anisotropic exchange
  if k is not None:
    hk = sm.geometry.get_hamiltonian(has_spin=False,fun=k) # initialize
    (mons,mhop) = c2h(hk.intra) # anisotropic part
    sm.hamiltonian.intra = sm.hamiltonian.intra + mons

  if sm.dimensionality == 0: pass # one dimensional
  elif sm.dimensionality == 1: # one dimensional
    (mons,mhop) = c2h(sm.hamiltonian.inter)
    (mons2,mhop2) = c2h(sm.hamiltonian.inter.H)
    sm.hamiltonian.inter = mhop
    sm.hamiltonian.intra += mons + mons2
  elif sm.dimensionality == 2: # one dimensional
    for name in ["tx","ty","txy","txmy"]: # loop over attributes
      (mons,mhop) = c2h(getattr(sm.hamiltonian,name))
      (mons2,mhop2) = c2h(getattr(sm.hamiltonian,name).H)
      setattr(sm.hamiltonian,name,mhop) # set hopping
      sm.hamiltonian.intra += mons + mons2 # add to onsite
  else: raise
  return sm
    



def sites2coupling(genij,spins):
  """Create the couplings of the HP-hamiltonian"""
  # SzSz terms
  # There are two matrices, one that renormalizes onsite energies
  # and another one that creates hoppings
  n = len(spins)
  ons = np.zeros((n,n),dtype=np.complex) # initial
  hop = np.zeros((n,n),dtype=np.complex) # initial
  iden = np.identity(n,dtype=np.complex) # identity
  for i in range(n): # loop over spins
    for j in range(n): # loop over spins
      cij = genij(i,j) # matrix with the couplings SiSj
      # the quantization axis is z
      ons[i,i] += cij[2,2] # Sz Sz component
      # there are missing terms!!!!!!
  # S+ S- terms
      sij = np.sqrt(spins[i]*spins[j]) # denominator
      hop[i,j] = -(cij[0,0] + cij[1,1])*sij/2.
  return (np.matrix(ons),np.matrix(hop)) # return matrices










def szsz(sm,fun=None):
  """Perform the Holfstein-Primakov for this spin model"""
  sm.hamiltonian = sm.geometry.get_hamiltonian(has_spin=False,fun=fun) # initialize
  spins = sm.spins # spins
  ns = len(spins) # number of sites
  inds = [i for i in range(ns)]
  def c2h(m0):
    def genij(i,j): return np.identity(3)*m0[i,j]
    return sites2coupling(genij,spins)
  # intra term
  for i in range(sm.hamiltonian.intra.shape[0]): 
    sm.hamiltonian.intra[i,i] = 0.0 # set coupling to zero
  (mons,mhop) = c2h(sm.hamiltonian.intra)
  mout = mons.copy() # copy matrix
  if sm.dimensionality == 0: pass # one dimensional
  elif sm.dimensionality == 1: # one dimensional
    (mons,mhop) = c2h(sm.hamiltonian.inter)
    (mons2,mhop2) = c2h(sm.hamiltonian.inter.H)
    mout += mons + mons2
  elif sm.dimensionality == 2: # one dimensional
    for name in ["tx","ty","txy","txmy"]: # loop over attributes
      (mons,mhop) = c2h(getattr(sm.hamiltonian,name))
      (mons2,mhop2) = c2h(getattr(sm.hamiltonian,name).H)
      mout += mons + mons2
  else: raise
  return mout  # return onsite matrix
    





def sites2coupling_sparse(mij,spins):
  """Create the couplings of the HP-hamiltonian"""
  # SzSz terms
  # There are two matrices, one that renormalizes onsite energies
  # and another one that creates hoppings
  n = len(spins)
  ons = coo_matrix(([],([],[])),shape=(n,n),dtype=np.complex)
  hop = coo_matrix(([],([],[])),shape=(n,n),dtype=np.complex)
  iden = sparseiden(n,dtype=np.complex) # identity matrix
  raise # not finished
  
  for i in range(n): # loop over spins
    for j in range(n): # loop over spins
      cij = genij(i,j) # matrix with the couplings SiSj
      # the quantization axis is z
      ons[i,i] += cij[2,2] # Sz Sz component
      # there are missing terms!!!!!!
  # S+ S- terms
      sij = np.sqrt(spins[i]*spins[j]) # denominator
      hop[i,j] = -(cij[0,0] + cij[1,1])*sij/2.
  return (np.matrix(ons),np.matrix(hop)) # return matrices












from .. import parallel
from .. import algebra
from ..klist import kmesh
import numpy as np
import scipy.sparse.linalg as slg

def get_eigenvectors(h,nk=10,kpoints=False,k=None,sparse=False,
        numw=None,energy=0.0):
  from scipy.sparse import csc_matrix as csc
  shape = h.intra.shape
  if numw is not None: sparse = True
  if h.dimensionality==0:
    if not sparse: vv = algebra.eigh(h.intra)
    if sparse: vv = slg.eigsh(csc(h.intra),k=numw,
            which="LM",sigma=energy,tol=1e-5)
    vecs = np.array([v for v in vv[1].transpose()])
    if kpoints: return vv[0],vecs,np.array([[0.,0.,0.] for e in vv[0]])
    else: return vv[0],vecs
  elif h.dimensionality>0:
    f = h.get_hk_gen()
    if k is None:
      kp = kmesh(h.dimensionality,nk=nk) # generate a mesh
    else:  kp = np.array([k]) # kpoint given on input
#    vvs = [lg.eigh(f(k)) for k in kp] # diagonalize k hamiltonian
    nkp = len(kp) # total number of k-points
    if sparse: # sparse Hamiltonians
        fk = lambda k: slg.eigsh(csc(f(k)),k=numw,which="LM",sigma=energy,tol=1e-5)
        vvs = parallel.pcall(fk,kp)
    else: # dense Hamiltonians
      if parallel.cores>1: # in parallel
#        vvs = parallel.multieigh([f(k) for k in kp]) # multidiagonalization
        vvs = parallel.pcall(lambda k: algebra.eigh(f(k)),kp)
      else: vvs = [algebra.eigh(f(k)) for k in kp] # 
    nume = sum([len(v[0]) for v in vvs]) # number of eigenvalues calculated
    eigvecs = np.zeros((nume,h.intra.shape[0]),dtype=np.complex) # eigenvectors
    eigvals = np.zeros(nume) # eigenvalues

    #### New way ####
#    eigvals = np.array([iv[0] for iv in vvs]).reshape(nkp*shape[0],order="F")
#    eigvecs = np.array([iv[1].transpose() for iv in vvs]).reshape((nkp*shape[0],shape[1]),order="F")
#    if kpoints: # return also the kpoints
#      kvectors = [] # empty list
#      for ik in kp: 
#        for i in range(h.intra.shape[0]): kvectors.append(ik) # store
#      return eigvals,eigvecs,kvectors
#    else:
#      return eigvals,eigvecs

    #### Old way, slightly slower but clearer ####
    iv = 0
    kvectors = [] # empty list
    for ik in range(len(kp)): # loop over kpoints
      vv = vvs[ik] # get eigenvalues and eigenvectors
      for (e,v) in zip(vv[0],vv[1].transpose()):
        eigvecs[iv] = v.copy()
        eigvals[iv] = e.copy()
        kvectors.append(kp[ik])
        iv += 1
    if kpoints: # return also the kpoints
#      for iik in range(len(kp)):
#        ik = kp[iik] # store kpoint 
#        for e in vvs[iik][0]: kvectors.append(ik) # store
      return eigvals,eigvecs,kvectors
    else:
      return eigvals,eigvecs
  else:
    raise





# library to deal with the spectral properties of the hamiltonian
import numpy as np
import scipy.linalg as lg
import scipy.sparse.linalg as slg
import os
from .operators import operator2list
from . import operators
from . import parallel
from . import kpm
from . import timing
from . import algebra
from . import densitymatrix

from .fermisurface import multi_fermi_surface

arpack_tol = 1e-5
arpack_maxiter = 10000

def fermi_surface(h,write=True,output_file="FERMI_MAP.OUT",
                    e=0.0,nk=50,nsuper=1,reciprocal=True,
                    delta=None,refine_delta=1.0,operator=None,
                    mode='eigen',num_waves=2,info=True):
    """Calculates the Fermi surface of a 2d system"""
    if operator is not None: # operator given
        operator = h.get_operator(operator) # get the operator
        if not operator.linear:
            if mode=="full": mode = "eigen"
    else: # no operator given
        if mode=="full":
            operator = np.matrix(np.identity(h.intra.shape[0]))
    if h.dimensionality!=2: raise  # continue if two dimensional
    hk_gen = h.get_hk_gen() # gets the function to generate h(k)
    kxs = np.linspace(-nsuper,nsuper,nk)  # generate kx
    kys = np.linspace(-nsuper,nsuper,nk)  # generate ky
    iden = np.identity(h.intra.shape[0],dtype=np.complex)
    kxout = []
    kyout = []
    if reciprocal: R = h.geometry.get_k2K_generator() # get function
    else:  R = lambda x: x
    # setup a reasonable value for delta
    if delta is None:  delta = 3./refine_delta*2./nk
    #### function to calculate the weight ###
    if mode=='full': # use full inversion
      def get_weight(hk,k=None):
        gf = algebra.inv((e+1j*delta)*iden - hk) # get green function
        gf = gf - algebra.inv((e-1j*delta)*iden - hk) # get green function
        if callable(operator): # callable operator
           tdos = -(operator(gf,k=k)).imag # get imaginary part
        else: tdos = -(operator*gf).imag # get imaginary part
        return np.trace(tdos).real # return trace
    elif mode=='eigen': # use full diagonalization
      def get_weight(hk,k=None):
        if operator is None:
            es = algebra.eigvalsh(hk)
            return np.sum(delta/((e-es)**2+delta**2)) # return weight
        else: # using an operator
            tmp,ds = h.get_dos(ks=[k],operator=operator,
                        energies=[e],delta=delta)
            return ds[0] # return weight
    elif mode=='lowest': # use sparse diagonalization
      def get_weight(hk,**kwargs):
        es,waves = slg.eigsh(hk,k=num_waves,sigma=e,tol=arpack_tol,which="LM",
                              maxiter = arpack_maxiter)
        return np.sum(delta/((e-es)**2+delta**2)) # return weight
    else: raise
  
  ##############################################
  
  
    # setup the operator
    rs = [] # empty list
    for x in kxs:
      for y in kxs:
        rs.append([x,y,0.]) # store
    def getf(r): # function to compute FS
        k = R(r)
        hk = hk_gen(k) # get hamiltonian
        return get_weight(hk,k=k)
    rs = np.array(rs) # transform into array
    from . import parallel
    kxout = rs[:,0] # x coordinate
    kyout = rs[:,1] # y coordinate
    if parallel.cores==1: # serial execution
        if info: ts = timing.Testimator(maxite=len(kxs)*len(kys),title="FS")
        kdos = [] # empty list
        for r in rs: # loop
          if info: ts.iterate()
          kdos.append(getf(r)) # add to the list
    else: # parallel execution
        kdos = parallel.pcall(getf,rs) # compute all
    if write:  # optionally, write in file
      f = open(output_file,"w") 
      for (x,y,d) in zip(kxout,kyout,kdos):
        f.write(str(x)+ "   "+str(y)+"   "+str(d)+"\n")
      f.close() # close the file
    return (kxout,kyout,d) # return result




def boolean_fermi_surface(h,write=True,output_file="BOOL_FERMI_MAP.OUT",
                    e=0.0,nk=50,nsuper=1,reciprocal=False,
                    delta=None):
    """Calculates the Fermi surface of a 2d system"""
    if h.dimensionality!=2: raise  # continue if two dimensional
    hk_gen = h.get_hk_gen() # gets the function to generate h(k)
    kxs = np.linspace(-nsuper,nsuper,nk)  # generate kx
    kys = np.linspace(-nsuper,nsuper,nk)  # generate ky
    kdos = [] # empty list
    kxout = []
    kyout = []
    if reciprocal: R = h.geometry.get_k2K() # get matrix
    # setup a reasonable value for delta
    if delta is None:
      delta = 8./np.max(np.abs(h.intra))/nk
    for x in kxs:
      for y in kxs:
        r = np.matrix([x,y,0.]).T # real space vectors
        k = np.array((R*r).T)[0] # change of basis
        hk = hk_gen(k) # get hamiltonian
        evals = lg.eigvalsh(hk) # diagonalize
        de = np.abs(evals - e) # difference with respect to fermi
        de = de[de<delta] # energies close to fermi
        if len(de)>0: kdos.append(1.0) # add to the list
        else: kdos.append(0.0) # add to the list
        kxout.append(x)
        kyout.append(y)
    if write:  # optionally, write in file
      f = open(output_file,"w") 
      for (x,y,d) in zip(kxout,kyout,kdos):
        f.write(str(x)+ "   "+str(y)+"   "+str(d)+"\n")
      f.close() # close the file
    return (kxout,kyout,d) # return result






















from .bandstructure import braket_wAw


def selected_bands2d(h,output_file="BANDS2D_",nindex=[-1,1],
               nk=50,nsuper=1,reciprocal=True,
               operator=None,k0=[0.,0.]):
  """ Calculate a selected bands in a 2d Hamiltonian"""
  if h.dimensionality!=2: raise  # continue if two dimensional
  hk_gen = h.get_hk_gen() # gets the function to generate h(k)
  kxs = np.linspace(-nsuper,nsuper,nk)+k0[0]  # generate kx
  kys = np.linspace(-nsuper,nsuper,nk)+k0[1]  # generate ky
  kdos = [] # empty list
  kxout = []
  kyout = []
  if reciprocal: R = h.geometry.get_k2K() # get matrix
  else:  R = np.array(np.identity(3)) # get identity
  # setup a reasonable value for delta
  # setup the operator
  operator = operator2list(operator) # convert into a list
  os.system("rm -f "+output_file+"*") # delete previous files
  fo = [open(output_file+"_"+str(i)+".OUT","w") for i in nindex] # files        
  for x in kxs:
    for y in kxs:
#      print("Doing",x,y)
      r = np.array([x,y,0.]) # real space vectors
      k = np.array(R)@r # change of basis
#      print(k)
      hk = hk_gen(k) # get hamiltonian
      if not h.is_sparse: evals,waves = lg.eigh(hk) # eigenvalues
      else: evals,waves = slg.eigsh(hk,k=max(np.abs(nindex))*2,sigma=0.0,
             tol=arpack_tol,which="LM") # eigenvalues
      waves = waves.transpose() # transpose
      epos,wfpos = [],[] # positive
      eneg,wfneg = [],[] # negative
      for (e,w) in zip(evals,waves): # loop
        if e>0.0: # positive
          epos.append(e)
          wfpos.append(w)
        else: # negative
          eneg.append(e)
          wfneg.append(w)
      # now sort the waves
      wfpos = [yy for (xx,yy) in sorted(zip(epos,wfpos))] 
      wfneg = [yy for (xx,yy) in sorted(zip(-np.array(eneg),wfneg))] 
      epos = sorted(epos)
      eneg = -np.array(sorted(-np.array(eneg)))
      for (i,j) in zip(nindex,range(len(nindex))): # loop over desired bands
        fo[j].write(str(x)+"     "+str(y)+"   ")
        if i>0: # positive
          fo[j].write(str(epos[i-1])+"  ")
          for op in operator: # loop over operators
            c = op.braket(wfpos[i-1]).real # expectation value
            fo[j].write(str(c)+"  ") # write in file
          fo[j].write("\n") # write in file
          
        if i<0: # negative
          fo[j].write(str(eneg[abs(i)-1])+"\n")
          for op in operator: # loop over operators
            c = op.braket(wfpos[abs(i)-1]).real # expectation value
            fo[j].write(str(c)+"  ") # write in file
          fo[j].write("\n") # write in file
  [f.close() for f in fo] # close file


get_bands = selected_bands2d




def ev2d(h,nk=50,nsuper=1,reciprocal=False,
               operator=None,k0=[0.,0.],kreverse=False):
  """ Calculate the expectation value of a certain operator"""
  if h.dimensionality!=2: raise  # continue if two dimensional
  hk_gen = h.get_hk_gen() # gets the function to generate h(k)
  kxs = np.linspace(-nsuper,nsuper,nk,endpoint=True)+k0[0]  # generate kx
  kys = np.linspace(-nsuper,nsuper,nk,endpoint=True)+k0[1]  # generate ky
  if kreverse: kxs,kys = -kxs,-kys
  kdos = [] # empty list
  kxout = []
  kyout = []
  if reciprocal: R = h.geometry.get_k2K() # get matrix
  else:  R = np.matrix(np.identity(3)) # get identity
  # setup the operator
  operator = operator2list(operator) # convert into a list
  fo = open("EV2D.OUT","w") # open file
  for x in kxs:
    for y in kxs:
      print("Doing",x,y)
      r = np.matrix([x,y,0.]).T # real space vectors
      k = np.array((R*r).T)[0] # change of basis
      hk = hk_gen(k) # get hamiltonian
      if not h.is_sparse: evals,waves = lg.eigh(hk) # eigenvalues
      else: evals,waves = slg.eigsh(hk,k=max(nindex)*2,sigma=0.0,
             tol=arpack_tol,which="LM") # eigenvalues
      waves = waves.transpose() # transpose
      eneg,wfneg = [],[] # negative
      for (e,w) in zip(evals,waves): # loop
        if e<0: # negative
          eneg.append(e)
          wfneg.append(w)
      fo.write(str(x)+"     "+str(y)+"   ") # write k-point
      for op in operator: # loop over operators
          c = sum([braket_wAw(w,op) for w in wfneg]).real # expectation value
          fo.write(str(c)+"  ") # write in file
      fo.write("\n") # write in file
  fo.close() # close file  





def ev(h,operator=None,nk=30,**kwargs):
  """Calculate the expectation value of a certain number of operators"""
  dm = densitymatrix.full_dm(h,nk=nk,use_fortran=True,**kwargs)
  if operator is None: # no operator given on input
    operator = [] # empty list
  elif not isinstance(operator,list): # if it is not a list
    operator = [operator] # convert to list
  out = [(dm@op).trace() for op in operator] 
  out = np.array(out) # return the result
  out = out.reshape(out.shape[0]) # reshape in case there are indexes
  return out # return array



def real_space_vev(h,operator=None,nk=1,nrep=3,name="REAL_SPACE_VEV.OUT",
        **kwargs):
    """Compute the expectation value in real space"""
    if nk>1: raise # only Gamma point implemented
    dm = densitymatrix.full_dm(h,nk=nk,**kwargs) # Gamma point DM
    operator = operators.Operator(operator) # convert to operator
    rho = operator(dm,k=[0.,0.,0.]) # compute the projected DM
    rho = np.diag(rho).real # extract the diagonal
    rho = h.full2profile(rho) # resum if necessary
    h.geometry.write_profile(rho,nrep=5,name=name)
    return rho










def total_energy(h,nk=10,nbands=None,use_kpm=False,random=False,
        kp=None,mode="mesh",tol=1e-1,fermi=0.0):
  """Return the total energy"""
  if nbands is None: h.turn_dense()
  if h.is_sparse and not use_kpm: 
      if nbands is None:
        print("Sparse Hamiltonian but no bands given, taking 20")
        nbands=20
  f = h.get_hk_gen() # get generator
  etot = 0.0 # initialize
  iv = 0
  def enek(k):
    """Compute energy in this kpoint"""
    hk = f(k)  # kdependent hamiltonian
    if use_kpm: # Kernel polynomial method
      return kpm.total_energy(hk,scale=10.,ntries=20,npol=100) # using KPM
    else: # conventional diagonalization
      if nbands is None: vv = algebra.eigvalsh(hk) # diagonalize k hamiltonian
      else: 
          vv,aa = slg.eigsh(hk,k=4*nbands,which="LM",sigma=0.0) 
          vv = -np.sort(-(vv[vv<fermi])) # negative eigenvalues
          vv = vv[0:nbands] # get the negative eigenvlaues closest to EF
      return np.sum(vv[vv<fermi]) # sum energies below fermi energy
  # compute energy using different modes
  if mode=="mesh":
    from .klist import kmesh
    kp = kmesh(h.dimensionality,nk=nk)
    etot = np.mean(parallel.pcall(enek,kp)) # compute total energy
  elif mode=="random":
    kp = [np.random.random(3) for i in range(nk)] # random points
    etot = np.mean(parallel.pcall(enek,kp)) # compute total eenrgy
  elif mode=="integrate":
    from scipy import integrate
    if h.dimensionality==1: # one dimensional
        etot = integrate.quad(enek,-1.,1.,epsabs=tol,epsrel=tol)[0]
    elif h.dimensionality==2: # two dimensional
        etot = integrate.dblquad(lambda x,y: enek([x,y]),-1.,1.,-1.,1.,
                epsabs=tol,epsrel=tol)[0]
    else: raise
  else: raise
  return etot






def eigenvalues(h0,nk=10,notime=True):
    """Return all the eigenvalues of a Hamiltonian"""
    from . import klist
    h = h0.copy() # copy hamiltonian
    h.turn_dense()
    ks = klist.kmesh(h.dimensionality,nk=nk) # get grid
    hkgen = h.get_hk_gen() # get generator
    if parallel.cores==1:
      es = [] # empty list
      if not notime: est = timing.Testimator(maxite=len(ks))
      for k in ks: # loop
          if not notime: est.iterate()
          es += algebra.eigvalsh(hkgen(k)).tolist() # add
    else:
        f = lambda k: algebra.eigvalsh(hkgen(k)) # add
        es = parallel.pcall(f,ks) # call in parallel
        es = np.array(es)
        es = es.reshape(es.shape[0]*es.shape[1])
    return es # return all the eigenvalues







def reciprocal_map(h,f,nk=40,reciprocal=True,nsuper=1,filename="MAP.OUT"):
  """ Calculates the reciprocal map of something"""
  if reciprocal: fR = h.geometry.get_k2K_generator()
  else: fR = lambda x: x
  fo = open(filename,"w") # open file
  nt = nk*nk # total number of points
  ik = 0
  ks = [] # list with kpoints
  from . import parallel
  for x in np.linspace(-nsuper,nsuper,nk,endpoint=False):
    for y in np.linspace(-nsuper,nsuper,nk,endpoint=False):
        ks.append([x,y,0.])
  tr = timing.Testimator(filename.replace(".OUT",""),maxite=len(ks))
  def fp(ki): # function to compute the quantity
      if parallel.cores == 1: tr.iterate()
      else: print("Doing",ki)
      k = fR(ki)
      return f(k) # call function
  bs = parallel.pcall(fp,ks) # compute all the Berry curvatures
  for (b,k) in zip(bs,ks): # write everything
      fo.write(str(k[0])+"   "+str(k[1])+"     "+str(b.real))
      fo.write("     "+str(b.imag)+"\n")
      fo.flush()
  fo.close() # close file



def singlet_map(h,nk=40,nsuper=3,mode="abs"):
    """Compute a map with the superconducting singlet pairing"""
    hk = h.get_hk_gen() # get function
    from .superconductivity import extract_pairing
    def f(k): # define function
      m = hk(k) # call Hamiltonian
      (uu,dd,ud) = extract_pairing(m) # extract the pairing
#      return np.abs(ud) # trace
#      return np.sum(np.abs(ud)) # trace
      if mode=="trace": return ud.trace()[0,0] # trace
      elif mode=="det": return np.linalg.det(ud) # trace
      elif mode=="abs": return np.sum(np.abs(ud)) # trace
    reciprocal_map(h,f,nk=nk,nsuper=nsuper,filename="SINGLET_MAP.OUT")



def pairing_map(h,**kwargs):
    """Compute a map with the superconducting singlet pairing"""
    h0 = h.copy()
    h0.remove_nambu()
    h0.setup_nambu_spinor()
    h = h - h0 # get only the pairing
    hk = h.get_hk_gen() # get function
    def f(k): # define function
        m = hk(k) # call Hamiltonian
        es = algebra.eigvalsh(m)
        return np.max(np.abs(es))
    reciprocal_map(h,f,filename="PAIRING_MAP.OUT",**kwargs)




def set_filling(h,filling=0.5,nk=10,extrae=0.,
    mode="ED",**kwargs):
    """
    Set the filling of a Hamiltonian
    - nk = 10, number of kpoints in each direction
    - filling = 0.5, filling of the lattice
    - extrae = 0.0, number of extra electrons
    """
    if h.has_eh: # quick workaround
        ef = h.get_fermi4filling(filling,nk=nk) # fermi energy
        h.add_onsite(-ef)
        return
    fill = filling + extrae/h.intra.shape[0] # filling
    n = h.intra.shape[0]
    use_kpm = False
    if n>algebra.maxsize: # use the KPM method
        mode="KPM"
        print("Using KPM in set_filling")
    if mode=="KPM": # use KPM
        es,ds = h.get_dos(energies=np.linspace(-5.0,5.0,1000),
                mode="KPM",nk=nk,**kwargs)
        from scipy.integrate import cumtrapz
        di = cumtrapz(ds,es)
        ei = (es[0:len(es)-1] + es[1:len(es)])/2.
        di /= di[len(di)-1] # normalize
        from scipy.interpolate import interp1d
        f = interp1d(di,ei) # interpolating function
        efermi = f(fill) # get the fermi energy
    elif mode=="ED": # dense Hamiltonian, use ED
        es = eigenvalues(h,nk=nk,notime=True)
        efermi = get_fermi_energy(es,fill)
    else: raise
    h.shift_fermi(-efermi) # shift the fermi energy


def get_fermi_energy(es,filling,fermi_shift=0.0):
  """Return the Fermi energy"""
  ne = len(es) ; ifermi = int(round(ne*filling)) # index for fermi
  if ifermi>=ne: ifermi = ne-1
  sorte = np.sort(es) # sorted eigenvalues
  if ifermi==0: return sorte[0] + fermi_shift
  fermi = (sorte[ifermi-1] + sorte[ifermi])/2.+fermi_shift # fermi energy
  return fermi


def get_fermi4filling(h,filling,nk=8):
    """Return the fermi energy for a certain filling"""
    if h.has_eh: 
        h0 = h.copy()
        h0.remove_nambu()
        return get_fermi4filling(h0,filling,nk=nk) # workaround
    else:
        es = eigenvalues(h,nk=nk,notime=True)
        return get_fermi_energy(es,filling)

def get_filling(h,**kwargs):
    """Get the filling of a Hamiltonian at this energy"""
    if h.check_mode("spinless_nambu"): # spinless Nambu Hamiltonian
        from .sctk import spinless
        return spinless.get_filling(h,**kwargs)
    elif h.check_mode("spinful_nambu"): raise # spinful Nambu
    else:
        es = eigenvalues(h,**kwargs) # eigenvalues
        es = np.array(es)
        esf = es[es<0.0]
        return len(esf)/len(es) # return filling





def eigenvalues_kmesh(h,nk=20):
    """Get the eigenvalues in a kmesh"""
    if h.dimensionality!=2: raise # only for 2d
    ne = h.intra.shape[0] # number of energies per k-point
    es = np.zeros((nk,nk,ne)) # array for the energies 
    hkgen = h.get_hk_gen() # get the generator
    kx = np.linspace(0.,1.,nk,endpoint=False)
    ky = np.linspace(0.,1.,nk,endpoint=False)
    for i in range(nk):
      ik = kx[i] # kx point   
      for j in range(nk):
        jk = ky[j] # ky point
        hk = hkgen([ik,jk]) # get the matrix
        ei = algebra.eigvalsh(hk) # get the energies
        es[i,j,:] = ei # store energies
    return es # return all the energies




def lowest_energies(h,n=4,k=None,**kwargs):
    """Return the lowest energy states in a k-point"""
    if k is None: raise
    es,ws = h.get_eigenvectors(kpoints=False,k=k,numw=2*n,**kwargs)
    es = [y for (x,y) in sorted(zip(np.abs(es),es))][0:n]
    es = np.sort(es)
    return es





from __future__ import print_function,division
import time

printall = False

class Testimator:
  def __init__(self,title="",maxite=None,silent=False):
    self.t0 = time.perf_counter() # starting time
    self.title = title
    self.maxite = maxite
    self.silent = silent
    self.i = 0
    if len(title)>0 and not silent: print(title)
  def remaining(self,i,tot):
    """Print the ramining time in this task"""
    t = time.perf_counter() # current time
    dt = t - self.t0 # difference in time
    out = self.title + " " # empty line
    for j in range(10):
      if j<(i/tot*10): out += "#"
      else: out += " "
    out += str(round(i/tot*100))+"% completed,"
    trem = dt/(i+1)*(tot-i) # remaining time
    out += " remaining time "+str(round(trem,1))+"s"
    out += ", total time "+str(round(dt,1))+"s"
    if not self.silent:
        if printall: print(out)
        else: print(out,end="\r")
  def iterate(self):
      if self.silent: return
      if self.maxite is not None: # of it has been provided
        self.remaining(self.i,self.maxite) # execute
        self.i += 1 # increase



from __future__ import print_function
from scipy.sparse import csc_matrix,bmat
from .rotate_spin import sx,sy,sz
import numpy as np
from numba import jit
from . import parallel

try:
  from . import kanemelef90
  use_fortran = True
except:
#  print("Not possible to use FORTRAN rutines, kanemele.py")
  use_fortran = False

import numbers 
def is_number(s):
    return isinstance(s, numbers.Number)

isnumber = is_number

def generalized_kane_mele(r1,r2,rm,fun=0.0,tol=1e-5):
  """Return the Kane-Mele generalized Hamiltonian"""
  if fun==0.0: return 0
  if is_number(fun): kmfun = lambda r: fun # function that always returns fun
  elif callable(fun): kmfun = fun # callable function
  else: raise # no idea
  nsites = len(r1) # number of sites
  mout = [[None for i in range(nsites)] for j in range(nsites)]
  for i in range(nsites):
    mout[i][i] = csc_matrix(np.zeros((2,2))) 
  for i in range(nsites): # loop over initial site
    for j in range(nsites): # loop over final site
      dr = r1[i]-r2[j] # difference
      if dr.dot(dr)>4.1:
        continue # if too far away, next iteration
      ur = km_vector(r1[i],r2[j],rm,tol=tol) # kane mele vector
      r3 = (r1[i] + r2[j])/2.0
      sm = (sx*ur[0] + sy*ur[1] + sz*ur[2])*kmfun(r3) # contribution
      if mout[i][j] is None: mout[i][j] = csc_matrix(1j*sm) # add contribution
      else: mout[i][j] += csc_matrix(1j*sm) # add contribution
  return bmat(mout) # return matrix


def km_vector(ri,rj,rm,use_fortran=use_fortran,
        tol=1e-5):
  """Return the Kane Mele vector"""
  if tol>1e-5: 
      use_fortran = False
  if use_fortran: return kanemelef90.kmvector(ri,rj,rm)
  else:  
      v = np.array([0.,0.,0.])
      return km_vector_jit(ri,rj,v,np.array(rm),tol=tol)
#      rmin = 1. -tol
#      rmax = 1. +tol
#      for k in range(len(rm)): # look for an intermediate site
#        dr1 = rm[k]-ri # difference
#        dr2 = rj-rm[k] # difference
#        if rmin<dr1.dot(dr1)<rmax and rmin<dr2.dot(dr2)<rmax: # if connected
#           ur = np.cross(dr1,dr2) # Kane Mele vector
#           return ur
#      return np.array([0.,0.,0.])


@jit(nopython=True)
def km_vector_jit(ri,rj,v,rm,tol=1e-5):
    rmin = 1. -tol
    rmax = 1. +tol
    for k in range(len(rm)): # look for an intermediate site
      dr1 = rm[k]-ri # difference
      dr2 = rj-rm[k] # difference
      if rmin<dr1.dot(dr1)<rmax and rmin<dr2.dot(dr2)<rmax: # if connected
         v = np.cross(dr1,dr2) # Kane Mele vector
         return v
    v = np.array([0.,0.,0.])
    return v





def haldane(r1,r2,rm,fun=0.0,sublattice=None):
  """Return the Haldane coupling"""
  if sublattice is None: sublattice = np.zeros(len(r1)) + 1.0
  if is_number(fun): 
      if fun==0.0: return 0 # skip
      kmfun = lambda r: fun # function that always returns fun
  elif callable(fun): kmfun = fun # callable function
  else: # anything else
      from .potentials import array2potential
      kmfun = array2potential(r1[:,0],r1[:,1],fun)
  nsites = len(r1) # number of sites
  mout = np.zeros((nsites,nsites),dtype=np.complex) # initialize
  from . import neighbor
  neighs = neighbor.connections(r1,rm) # list with neighbors of each site
  for i in range(nsites): # loop over initial site
    rijs = [rm[kk] for kk in neighs[i]] # loop over first neighbors
    if len(rijs)==0: continue
    for j in range(nsites): # loop over final site
      dr = r1[i]-r2[j] # difference
      if dr.dot(dr)>4.1:
        continue # if too far away, next iteration
#      ur = km_vector(r1[i],r2[j],rm) # kane mele vector
#      print(rijs)
      ur = km_vector(r1[i],r2[j],rijs) # kane mele vector
      r3 = (r1[i] + r2[j])/2.0
      sm = ur[2]*kmfun(r3) # clockwise or anticlockwise
      mout[i,j] = 1j*sm*(sublattice[i]+sublattice[j])/2. # store
  return csc_matrix(mout) # return matrix



def get_haldane_function(g,stagger=False):
    """Return a function that computes the Haldane coupling"""
    rm = g.multireplicas(3)
    def f(r1,r2):
      m = haldane([r1],[r2],rm,fun=1.0).todense()[0,0]
      if not stagger: return m # conventional Haldane
      i = g.get_index(r1,replicas=True)
      j = g.get_index(r2,replicas=True)
      i = g.sublattice[i]
      j = g.sublattice[j]
      m = m*(i+j)/2
      return m
    return f # return function











def add_haldane(h,t):
  """Add Haldane to a Hamiltonian"""
#  if h.has_spin: raise # not for spinful
#  if h.has_eh: raise # not for spinful
  add_haldane_like(h,t,haldane,sublattice=None) # add Haldane to a Hamiltonian


def add_modified_haldane(h,t):
  """Add Haldane to a Hamiltonian"""
#  if h.has_eh: raise # not for spinful
  if not h.geometry.has_sublattice: return # if it does not have sublattice
  add_haldane_like(h,t,haldane,sublattice=h.geometry.sublattice) 


def add_anti_kane_mele(h,t):
  """Add Haldane to a Hamiltonian"""
  if not h.has_spin: raise
  if not h.geometry.has_sublattice: return # if it does not have sublattice
  add_haldane_like(h,t,haldane,sublattice=h.geometry.sublattice,
          time_reversal = True) 


def add_kane_mele(self,t,**kwargs):
  """Add to a Hamiltonian a Haldane-like hopping"""
  if not self.has_spin: self.turn_spinful() # spilful Hamiltonian
  from .multicell import close_enough # check if two rs are close
  g = self.geometry
  if not self.has_spin: raise  # only for spinfull
  if self.is_multicell:   # multicell Hamiltonian
    ncells = 2 # number of neighboring cells to check
    if self.dimensionality==0: rs = g.r # fix for zero dimensional
    elif self.dimensionality==1:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        rtmp = [ri + i*g.a1 for ri in g.r] # new positions
        if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
          rs += rtmp
    elif self.dimensionality==2:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        for j in range(-ncells,ncells+1):
          rtmp = [ri + i*g.a1 +j*g.a2 for ri in g.r] # new positions
          if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
            rs += rtmp
    elif self.dimensionality==3:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        for j in range(-ncells,ncells+1):
          for k in range(-ncells,ncells+1):
            rtmp = [ri + i*g.a1 +j*g.a2 + k*g.a3 for ri in g.r] # new positions
            if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
              rs += rtmp
    else: raise

    m = generalized_kane_mele(g.r,g.r,rs,fun=t,**kwargs) # kane mele coupling
    m = self.spinful2full(m) # convert the matrix
    self.intra = self.intra + m # add matrix
    for i in range(len(self.hopping)): # loop over hoppings
      d = self.hopping[i].dir
#      print("Generating Haldane-like",d,end="\r")
      r2 = [ri + d[0]*g.a1 + d[1]*g.a2 +d[2]*g.a3 for ri in g.r] # second vectors
      m = generalized_kane_mele(g.r,r2,rs,fun=t) # kane mele coupling
      m = self.spinful2full(m) # convert the matrix
      self.hopping[i].m += m
    return

  else:  # conventional Hamiltonian
    if self.dimensionality==0:  # zero dimensional
      rs = g.r # positions
    elif self.dimensionality==1:  # one dimensional
      rs = [] # all the cells
      for i in (-1,0,1):
        for ri in g.r:
          rs.append(ri + i*g.a1)
    elif self.dimensionality==2:  # two dimensional
      rs = [] # all the cells
      for i in (-1,0,1): # loop over neighbouring cells
        for j in (-1,0,1):
          for ri in g.r:
            rs.append(ri + i*g.a1 + j*g.a2)
    else: raise
    # now create the hamiltonian
    m = generalized_kane_mele(g.r,g.r,rs,fun=t,**kwargs) # kane mele coupling
    m = self.spinful2full(m) # convert the matrix
    self.intra = self.intra + m
    if self.dimensionality==0: pass  # zero dimensional
    elif self.dimensionality==1:  # zero dimensional
      r2 = [ri + g.a1 for ri in g.r] # new positions
      m = generalized_kane_mele(g.r,r2,rs,fun=t,**kwargs)
      m = self.spinful2full(m) # convert the matrix
      self.inter = self.inter + m 
    elif self.dimensionality==2:  # two dimensional
      r2 = [ri + 1*g.a1 + 0*g.a2 for ri in g.r] # second vectors
      m = generalized_kane_mele(g.r,r2,rs,fun=t,**kwargs)
      m = self.spinful2full(m) # convert the matrix
      self.tx = self.tx + m 
      ###############
      r2 = [ri + 0*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      m = generalized_kane_mele(g.r,r2,rs,fun=t,**kwargs)
      m = self.spinful2full(m) # convert the matrix
      self.ty = self.ty + m 
      #################
      r2 = [ri + 1*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      m = generalized_kane_mele(g.r,r2,rs,fun=t)
      m = self.spinful2full(m) # convert the matrix
      self.txy = self.txy + m 
      #################
      r2 = [ri + 1*g.a1  -1*g.a2 for ri in g.r] # second vectors
      m = generalized_kane_mele(g.r,r2,rs,fun=t)
      m = self.spinful2full(m) # convert the matrix
      self.txmy = self.txmy + m
    else: raise



from .increase_hilbert import spinful,spinful_sparse



def add_haldane_like(self,t,spinless_generator,
        sublattice=None,time_reversal=False):
  """Add to a Hamiltonian a Haldane-like hopping"""
  if isnumber(t): 
      if t==0.0: return # skip
  g = self.geometry
  if sublattice is None: sublattice = np.zeros(len(g.r)) + 1.0 # ones
  def generator(r1,r2,rs,fun=t,sublattice=sublattice): # define function
     m = spinless_generator(r1,r2,rs,fun=t,
             sublattice=sublattice)
     return self.spinless2full(m,time_reversal=time_reversal) # return matrix
#     if self.has_spin: return spinful_sparse(m) # spinful
#     else: return m # spinless
  from .multicell import close_enough # check if two rs are close
  if self.is_multicell:   # multicell Hamiltonians
    if self.dimensionality==0:  rs = g.r
    else: # higher dimensional
      ncells = self.geometry.get_ncells() # number of unit cells
      if self.dimensionality==1:  # three dimensional
        rs = [] # all the cells
        for i in range(-ncells,ncells+1): # loop over neighbouring cells
          rtmp = self.geometry.replicas(d=[i,0,0]) 
          if close_enough(g.r,rtmp,rcut=2.1): # if these positions are not too far
            rtmp = [ri for ri in rtmp]
            rs += rtmp
      elif self.dimensionality==2:  # three dimensional
        rs = [] # all the cells
        for i in range(-ncells,ncells+1): # loop over neighbouring cells
          for j in range(-ncells,ncells+1):
            rtmp = self.geometry.replicas(d=[i,j,0]) 
            if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
              rtmp = [ri for ri in rtmp]
              rs += rtmp
      elif self.dimensionality==3:  # three dimensional
        rs = [] # all the cells
        for i in range(-ncells,ncells+1): # loop over neighbouring cells
          for j in range(-ncells,ncells+1):
            for k in range(-ncells,ncells+1):
              rtmp = self.geometry.replicas(d=[i,j,k]) 
              if close_enough(g.r,rtmp,rcut=2.1): # if not too far
                rtmp = [ri for ri in rtmp]
                rs += rtmp
      else: raise

#    self.intra += generator(g.r,g.r,rs,fun=t,sublattice=sublattice) # coupling
    dirs = [[0,0,0]] + [t.dir for t in self.hopping] # directions
    def pfun(d): # function to parallelize
#    for i in range(len(self.hopping)): # loop over hoppings
#      print("Generating Haldane-like",d)
      r2 = self.geometry.replicas(d=d)
      return generator(g.r,r2,rs,fun=t,sublattice=sublattice) 
    ms = parallel.pcall(pfun,dirs) # get matrices
    self.intra = self.intra + ms[0] # intracell matrix
    for i in range(len(self.hopping)):
        self.hopping[i].m += ms[i+1] # store
    return

  else:  # conventional Hamiltonian
    if self.dimensionality==0:  # zero dimensional
      rs = g.r # positions
    elif self.dimensionality==1:  # one dimensional
      rs = [] # all the cells
      for i in (-1,0,1):
        for ri in g.r:
          rs.append(ri + i*g.a1)
    elif self.dimensionality==2:  # two dimensional
      rs = [] # all the cells
      for i in (-1,0,1): # loop over neighbouring cells
        for j in (-1,0,1):
          for ri in g.r:
            rs.append(ri + i*g.a1 + j*g.a2)
    else: raise
    # now create the hamiltonian
    self.intra = self.intra + generator(g.r,g.r,rs,fun=t,sublattice=sublattice) # kane mele coupling
    if self.dimensionality==0: pass  # zero dimensional
    elif self.dimensionality==1:  # zero dimensional
      r2 = [ri + g.a1 for ri in g.r] # new positions
      self.inter += generator(g.r,r2,rs,fun=t,sublattice=sublattice) # kane mele coupling
    elif self.dimensionality==2:  # two dimensional
      r2 = [ri + 1*g.a1 + 0*g.a2 for ri in g.r] # second vectors
      self.tx = self.tx + generator(g.r,r2,rs,fun=t,sublattice=sublattice)
      r2 = [ri + 0*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      self.ty = self.ty + generator(g.r,r2,rs,fun=t,sublattice=sublattice)
      r2 = [ri + 1*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      self.txy = self.txy + generator(g.r,r2,rs,fun=t,sublattice=sublattice)
      r2 = [ri + 1*g.a1  -1*g.a2 for ri in g.r] # second vectors
      self.txmy = self.txmy + generator(g.r,r2,rs,fun=t,sublattice=sublattice)
    else: raise




def add_kane_mele_old(self,t):
  """Add to a Hamiltonian a Haldane-like hopping"""
#  print("This does not work with eh, kanemele")
  if not self.has_spin: self.turn_spinful() # spilful Hamiltonian
  from .multicell import close_enough # check if two rs are close
  g = self.geometry
  if not self.has_spin: raise  # only for spinfull
  if self.is_multicell:   # multicell Hamiltonians
    ncells = 4 # number of neighboring cells to check
    if self.dimensionality==1:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        rtmp = [ri + i*g.a1 for ri in g.r] # new positions
        if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
          rs += rtmp
    elif self.dimensionality==2:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        for j in range(-ncells,ncells+1):
          rtmp = [ri + i*g.a1 +j*g.a2 for ri in g.r] # new positions
          if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
            rs += rtmp
    elif self.dimensionality==3:  # three dimensional
      rs = [] # all the cells
      for i in range(-ncells,ncells+1): # loop over neighbouring cells
        for j in range(-ncells,ncells+1):
          for k in range(-ncells,ncells+1):
            rtmp = [ri + i*g.a1 +j*g.a2 + k*g.a3 for ri in g.r] # new positions
            if close_enough(g.r,rtmp,rcut=2.1): # if this positions are not too far
              rs += rtmp
    else: raise

    self.intra += generalized_kane_mele(g.r,g.r,rs,fun=t) # kane mele coupling
    for i in range(len(self.hopping)): # loop over hoppings
      d = self.hopping[i].dir
#      print("Adding Kane Mele in hopping",d,end="\r")
      r2 = [ri + d[0]*g.a1 + d[1]*g.a2 +d[2]*g.a3 for ri in g.r] # second vectors
      self.hopping[i].m += generalized_kane_mele(g.r,r2,rs,fun=t) # kane mele coupling
    return

  else:  # conventional Hamiltonian
    if self.dimensionality==0:  # zero dimensional
      rs = g.r # positions
    elif self.dimensionality==1:  # one dimensional
      rs = [] # all the cells
      for i in (-1,0,1):
        for ri in g.r:
          rs.append(ri + i*g.a1)
    elif self.dimensionality==2:  # two dimensional
      rs = [] # all the cells
      for i in (-1,0,1): # loop over neighbouring cells
        for j in (-1,0,1):
          for ri in g.r:
            rs.append(ri + i*g.a1 + j*g.a2)
    else: raise
    # now create the hamiltonian
    self.intra = self.intra + generalized_kane_mele(g.r,g.r,rs,fun=t) # kane mele coupling
    if self.dimensionality==0: pass  # zero dimensional
    elif self.dimensionality==1:  # zero dimensional
      r2 = [ri + g.a1 for ri in g.r] # new positions
      self.inter += generalized_kane_mele(g.r,r2,rs,fun=t) # kane mele coupling
    elif self.dimensionality==2:  # two dimensional
      r2 = [ri + 1*g.a1 + 0*g.a2 for ri in g.r] # second vectors
      self.tx = self.tx + generalized_kane_mele(g.r,r2,rs,fun=t)
      r2 = [ri + 0*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      self.ty = self.ty + generalized_kane_mele(g.r,r2,rs,fun=t)
      r2 = [ri + 1*g.a1 + 1*g.a2 for ri in g.r] # second vectors
      self.txy = self.txy + generalized_kane_mele(g.r,r2,rs,fun=t)
      r2 = [ri + 1*g.a1  -1*g.a2 for ri in g.r] # second vectors
      self.txmy = self.txmy + generalized_kane_mele(g.r,r2,rs,fun=t)
    else: raise



# library to deal with the spectral properties of the hamiltonian
import numpy as np
import scipy.linalg as lg
import scipy.sparse.linalg as slg
import os
from .operators import operator2list
from . import parallel
from . import kpm
from . import timing
from . import algebra
from . import filesystem as fs

arpack_tol = 1e-5
arpack_maxiter = 10000

def fermi_surface_generator(h,
                    energies=[0.0],nk=50,nsuper=1,reciprocal=True,
                    delta=1e-2,refine_delta=1.0,operator=None,
                    full_bz=False, # special flag for unfolding
                    numw=20,info=True):
  """Calculates the Fermi surface of a 2d system"""
  if h.is_sparse: mode = "sparse"
  else: mode = "full"
  energies = np.array(energies) # convert to array
  if h.dimensionality!=2: raise  # continue if two dimensional
  hk_gen = h.get_hk_gen() # gets the function to generate h(k)
  if full_bz:
      kxs = np.linspace(0.,nsuper,nk,endpoint=True)  # generate kx
      kys = np.linspace(0.,nsuper,nk,endpoint=True)  # generate ky
  else:
      kxs = np.linspace(-nsuper,nsuper,nk,endpoint=True)  # generate kx
      kys = np.linspace(-nsuper,nsuper,nk,endpoint=True)  # generate ky
  iden = np.identity(h.intra.shape[0],dtype=np.complex)
  kxout = []
  kyout = []
  if reciprocal: fR = h.geometry.get_k2K_generator() # get matrix
  else:  fR = lambda x: x # get identity
  # setup a reasonable value for delta
  #### function to calculate the weight ###
  def get_weight(hk,k=None):
      if operator is None:
          if mode=='full': es = algebra.eigvalsh(hk) # get eigenvalues
          elif mode=='sparse': es = algebra.smalleig(hk,numw=numw)
          ws = [np.sum(delta/((e-es)**2+delta**2)) for e in energies] # weights
          return np.array(ws) # return weights
      else:
          tmp,ds = h.get_dos(ks=[k],operator=operator,
                        energies=energies,delta=delta)
          return ds # return weight
##############################################
  ts = timing.Testimator()
  # setup the operator
  rs = [] # empty list
  for x in kxs:
    for y in kys:
      rs.append([x,y,0.]) # store
  def getf(r): # function to compute FS
      k = fR(r) # get the reciprocal space vector
      hk = hk_gen(k) # get hamiltonian
      return get_weight(hk,k=k) # get the array with the weights
  rs = np.array(rs) # transform into array
  from . import parallel
  kxout = rs[:,0] # x coordinate
  kyout = rs[:,1] # y coordinate
  if parallel.cores==1: # serial execution
      kdos = [] # empty list
      for r in rs: # loop
        if info: print("Doing",r)
        kdos.append(getf(r)) # add to the list
  else: # parallel execution
      kdos = parallel.pcall(getf,rs) # compute all
  kdos = np.array(kdos) # transform into an array
  return energies,rs,kdos


def multi_fermi_surface(h,nk=50,delta=1e-2,
        output_folder="MULTIFERMISURFACE",**kwargs):
    """Compute several fermi surfaces"""
    energies,rs,kdos = fermi_surface_generator(h,nk=nk,delta=delta,**kwargs)
    fs.rmdir(output_folder) # remove folder
    fs.mkdir(output_folder) # create folder
    fo = open(output_folder+"/"+output_folder+".TXT","w")
    for i in range(len(energies)): # loop over energies
        filename = output_folder+"_"+str(energies[i])+"_.OUT" # name
        name = output_folder+"/"+filename
        np.savetxt(name,np.array([rs[:,0],rs[:,1],kdos[:,i]]).T)
        fo.write(filename+"\n")
    name = output_folder+"/DOS.OUT"
    name = "DOS.OUT"
    np.savetxt(name,np.array([energies,np.sum(kdos,axis=0)]).T)
    fo.close()





from __future__ import print_function

import numpy as np
from . import geometry
from .hamiltonians import sx,sy,sz
from scipy.sparse import csc_matrix

def hamiltonian(mw=1.0):
  """Return the Hamiltonian for the surface of a 3d TI"""
  g = geometry.single_square_lattice() # get the geometry
  return geometry2hamiltonian(g,mw=mw)

  h = g.get_hamiltonian() # initialize the Hamiltonian
  h.tx = 1j*sy.todense() # update hopping 
  h.ty = -1j*sx.todense() # update hopping 
  # Now add the Wilson mass term
  h.intra = 2*mw*sz.todense() # update hopping 
  h.tx -= mw/2.*sz.todense() # update hopping 
  h.ty -= mw/2.*sz.todense() # update hopping 
  return h # return the Hamiltonian


def geometry2hamiltonian(g,mw=0.6,is_sparse=False):
  """Given an input geometry, generate the Hamiltonian"""


  ## Only valid for square lattice ###
  tol = 0.001 # tolerancy
  ons = csc_matrix(mw*np.matrix(np.identity(2,dtype=np.complex))) # onsite
  ons = csc_matrix(2.*sz*mw) # onsite
  tx = csc_matrix(1j*sy-mw/2.*sz) # hopping in x direction
  ty = csc_matrix(-1j*sx-mw/2.*sz) # hopping in y direction
  dx = np.array([1.,0.,0.])
  dy = np.array([0.,1.,0.])
  def f(r1,r2): # function the returns the hopping
    dr = r2-r1 # distance
    if dr.dot(dr)<tol: return ons # onsite
    elif (dr-dx).dot(dr-dx)<tol: return tx # onsite
    elif (dr-dy).dot(dr-dy)<tol: return ty # onsite
    elif (dr+dx).dot(dr+dx)<tol: return tx.H # onsite
    elif (dr+dy).dot(dr+dy)<tol: return ty.H # onsite
    else: return ons*0.

  h = g.get_hamiltonian(fun=f,spinful_generator=True) # generate 
  return h
  





from __future__ import print_function,division
import numpy as np
from . import green
from . import dos
from . import klist
from . import kpm
from . import timing
from . import multicell
from . import kpm
from . import sculpt
from . import parallel
from . import algebra

def write_kdos(k=0.,es=[],ds=[],new=True):
  """ Write KDOS in a file"""
  if new: f = open("KDOS.OUT","w") # open new file
  else: f = open("KDOS.OUT","a") # continue writting
  for (e,d) in zip(es,ds): # loop over e and dos
    f.write(str(k)+"     ")
    f.write(str(e)+"     ")
    f.write(str(d)+"\n")
  f.close()





def kdos1d_sites(h,sites=[0],scale=10.,nk=100,npol=100,kshift=0.,
                  ewindow=None,info=False):
  """ Calculate kresolved density of states of
  a 1d system for a certain orbitals"""
  if h.dimensionality!=1: raise # only for 1d
  ks = np.linspace(0.,1.,nk) # number of kpoints
  h.turn_sparse() # turn the hamiltonian sparse
  hkgen = h.get_hk_gen() # get generator
  if ewindow is None:  xs = np.linspace(-0.9,0.9,nk) # x points
  else:  xs = np.linspace(-ewindow/scale,ewindow/scale,nk) # x points
  write_kdos() # initialize file
  for k in ks: # loop over kpoints
    mus = np.array([0.0j for i in range(2*npol)]) # initialize polynomials
    hk = hkgen(k+kshift) # hamiltonian
    for isite in sites:
      mus += kpm.local_dos(hk/scale,i=isite,n=npol)
    ys = kpm.generate_profile(mus,xs) # generate the profile
    write_kdos(k,xs*scale,ys,new=False) # write in file (append)
    if info: print("Done",k)

#
#def surface(h,energies=None,klist=None,delta=0.01):
#  """Return bulk and surface DOS"""
#  bout = [] # empty list, bulk
#  sout = [] # empty list, surface
#  for k in klist:
#    for energy in energies:
#      gs,sf = green.green_kchain(h,k=k,energy=energy,delta=delta,only_bulk=False) 
#      bout.append(gs.trace()[0,0].imag) # bulk
#      sout.append(sf.trace()[0,0].imag) # surface
#  bout = np.array(bout).reshape((len(energies),len(klist))) # convert to array
#  sout = np.array(sout).reshape((len(energies),len(klist))) # convert to array
#  return (bout.transpose(),sout.transpose())
#
#
#






def write_surface(h,energies=np.linspace(-.5,.5,300),
        klist=None,delta=None,operator=None,hs=None):
  if delta is None: delta = (np.max(energies)-np.min(energies))/len(energies)
  if h.dimensionality==1:
    write_surface_1d(h,energies=energies,delta=delta,
                         operator=operator)
  elif h.dimensionality==2:
    write_surface_2d(h,energies=energies,klist=klist,delta=delta,
                         operator=operator,hs=hs)
  elif h.dimensionality==3:
    write_surface_3d(h,energies=energies,klist=klist,delta=delta)
  else: raise




def write_surface_1d(h,energies=None,delta=None,
        operator=None):
  if energies is None: energies = np.linspace(-.5,.5,200)
  if delta is None: delta = (max(energies)-min(energies))/len(energies)
  h = h.get_no_multicell()
  fo  = open("SURFACE_DOS.OUT","w") # open file
  for energy in energies:
      gs,sf = green.green_renormalization(h.intra,h.inter,
              energy=energy,delta=delta) # surface green function 
      if operator is None: op = np.identity(h.intra.shape[0]) # identity matrix
      elif callable(operator): op = callable(op)
      else: op = operator # assume a matrix
      db = -algebra.trace(gs*op).imag # bulk
      ds = -algebra.trace(sf*op).imag # surface
      fo.write(str(energy)+"   "+str(ds)+"   "+str(db)+"\n")
      fo.flush()
  fo.close()





def write_surface_2d(h,energies=None,klist=None,delta=0.01,
                         operator=None,hs=None):
  bout = [] # empty list, bulk
  sout = [] # empty list, surface
  if klist is None: 
      klist = [[i,0.,0.] for i in np.linspace(-.5,.5,50)]
  if energies is None: energies = np.linspace(-.5,.5,50)
  fo  = open("KDOS.OUT","w") # open file
  for k in klist:
    print("Doing k-point",k)
    for energy in energies:
      gs,sf = green.green_kchain(h,k=k,energy=energy,delta=delta,
                       only_bulk=False,hs=hs) # surface green function 
      if operator is None: op = np.identity(h.intra.shape[0]) # identity matrix
      elif callable(operator): op = callable(op)
      else: op = operator # assume a matrix
      db = -algebra.trace(gs*op).imag # bulk
      ds = -algebra.trace(sf*op).imag # surface
      fo.write(str(k[0])+"   "+str(energy)+"   "+str(ds)+"   "+str(db)+"\n")
      fo.flush()
  fo.close()


def write_surface_3d(h,energies=None,klist=None,delta=0.01):
  raise # not implemented
  if h.dimensionality != 3: raise # only for 3d
  ho = h.copy() # copy Hamiltonian
  ho = ho.turn_multicell() # multicell Hamiltonian
  bout = [] # empty list, bulk
  sout = [] # empty list, surface
  if klist is None: raise
  if energies is None: energies = np.linspace(-.5,.5,50)
  fo  = open("KDOS.OUT","w") # open file
  for k in klist:
    for energy in energies:
      gs,sf = green.green_kchain(h,k=k,energy=energy,delta=delta,only_bulk=False) 
      db = -algebra.trace(gs).imag # bulk
      ds = -algebra.trace(sf).imag # surface
      fo.write(str(k)+"   "+str(energy)+"   "+str(ds)+"   "+str(db)+"\n")



def kdos_bands(h,use_kpm=False,kpath=None,scale=10.0,frand=None,
                 P = None,
                 ewindow=4.0,ne=1000,delta=0.01,ntries=10,nk=100,
                 operator=None,energies=np.linspace(-3.0,3.0,200),
                 mode="ED",**kwargs):
  """Calculate the KDOS bands using the KPM"""
  if use_kpm: mode ="KPM" # conventional method
  if mode=="ED":
      from . import dos
      def pfun(k):
        (es,ds) = h.get_dos(ks=[k],operator=operator,energies=energies,
                delta=delta)
        return energies,ds
  elif mode=="green":
    f = h.get_gk_gen(delta=delta) # Green generator
    def pfun(k): # do it for this k-point
        def gfun(e):
            m = f(k=k,e=e) # Green's function
            m = green.GtimesO(m,operator,k=k)
            return -algebra.trace(m).imag # return DOS
        return energies,np.array([gfun(e) for e in energies])
  elif mode=="KPM": # KPM method
    if operator is not None: 
        from .operators import Operator
        operator = Operator(operator).get_matrix() # convert to matrix
    hkgen = h.get_hk_gen() # get generator
    def pfun(k): # do it for this k-point
      print("Doing",k)
      hk = hkgen(k) # get Hamiltonian
      npol = int(scale/delta) # number of polynomials
      (x,y) = kpm.pdos(hk,scale=scale,npol=npol,ne=ne,P=P,
                   operator=operator,
                   ewindow=ewindow,ntries=ntries,x=energies) # compute
      return (x,y)
  kpath = h.geometry.get_kpath(kpath,nk=nk) # generate kpath
  ### Now compute and write in a file
  ik = 0
  out = parallel.pcall(pfun,kpath) # compute all
  fo = open("KDOS_BANDS.OUT","w") # open file
  for k in kpath: # loop over kpoints
    (x,y) = out[ik] # get this one
    for (ix,iy) in zip(x,y): # loop
      fo.write(str(ik/len(kpath))+"   ")
      fo.write(str(ix)+"   ")
      fo.write(str(iy)+"\n")
    fo.flush()
    ik += 1
  fo.close()









def write_surface_kpm(h,ne=400,klist=None,scale=4.,npol=200,w=20,ntries=20):
  """Write the surface DOS using the KPM"""
  if klist is None: klist = np.linspace(-.5,.5,50)
  fo  = open("KDOS.OUT","w") # open file
  for k in klist:
    print("Doing kpoint",k)
    if h.dimensionality==2: 
      (intra,inter) = h.kchain(k) # k hamiltonian
      (es,ds,dsb) = kpm.edge_dos(intra,inter,scale=scale,w=w,npol=npol,
                            ne=ne,bulk=True)
    # if the Hamiltonian is 1d from the beginning
    elif h.dimensionality==1: 
      intra,inter = h.intra,h.inter # 1d hamiltonian
      dd = h.intra.shape[0] # dimension
      inde = np.zeros(dd) # array with zeros
      indb = np.zeros(dd) # array with zeros
      for i in range(dd//10): # one tenth
        inde[i] = 1. # use this one
        indb[4*dd//10 + i] = 1. # use this one
      def gedge(): return (np.random.random(len(inde))-0.5)*inde
      def gbulk(): return (np.random.random(len(indb))-0.5)*(indb)
      # hamiltonian
      h0 = intra + inter*np.exp(1j*np.pi*2.*k) + (inter*np.exp(1j*np.pi*2.*k)).H
      xs = np.linspace(-0.9,0.9,4*npol) # x points
      es = xs*scale
      # calculate the bulk
      mus = kpm.random_trace(h0/scale,ntries=ntries,n=npol,fun=gbulk)
      dsb = kpm.generate_profile(mus,xs) # generate the profile
      # calculate the edge
      mus = kpm.random_trace(h0/scale,ntries=ntries,n=npol,fun=gedge)
      ds = kpm.generate_profile(mus,xs) # generate the profile
    else: raise
    for (e,d1,d2) in zip(es,ds,dsb):
      fo.write(str(k)+"   "+str(e)+"   "+str(d1)+"    "+str(d2)+"\n")
  fo.close()




def interface(h1,h2,energies=np.linspace(-1.,1.,100),operator=None,
                    delta=None,kpath=None,dh1=None,dh2=None,nk=50):
  """Get the surface DOS of an interface"""
  from scipy.sparse import csc_matrix,bmat
  if delta is None:
      delta = 1*(max(energies) - min(energies))/len(energies)
  if kpath is None: 
    if h1.dimensionality==1:
      kpath = [[0.,0.,0.]]
    elif h1.dimensionality==3:
      g2d = h1.geometry.copy() # copy Hamiltonian
      g2d = sculpt.set_xy_plane(g2d)
      kpath = klist.default(g2d,nk=nk)
    elif h1.dimensionality==2:
      kpath = [[k,0.,0.] for k in np.linspace(0.,1.,nk)]
    else: raise
#  tr = timing.Testimator("KDOS") # generate object
#  tr.remaining(ik,len(kpath)) # generate object
  ik = 0
  h1 = h1.get_multicell() # multicell Hamiltonian
  h2 = h2.get_multicell() # multicell Hamiltonian
  def computek(ik):
    k = kpath[ik] # get this one
#    for energy in energies:
#  (b1,s1,b2,s2,b12) = green.interface(h1,h2,k=k,energy=energy,delta=delta)
#      out = green.interface(h1,h2,k=k,energy=energy,delta=delta)
    outs = green.interface_multienergy(h1,h2,k=k,energies=energies,
            delta=delta,dh1=dh1,dh2=dh2)
    outstr = ""
    for (energy,out) in zip(energies,outs):
      if operator is None: 
        op = np.identity(h1.intra.shape[0]*2) # normal cell
        ops = np.identity(h1.intra.shape[0]) # supercell 
#      elif callable(operator): op = callable(op)
      else:
        op = operator # normal cell 
        ops = bmat([[csc_matrix(operator),None],[None,csc_matrix(operator)]])
      # write everything
      outstr += str(ik)+"   "+str(energy)+"   "
      for g in out: # loop
        if g.shape[0]==op.shape[0]: d = -algebra.trace(g@op).imag # bulk
        else: d = -algebra.trace(g@ops).imag # interface
        outstr += str(d)+"   "
      outstr += "\n"
    return outstr
  out = parallel.pcall(computek,range(len(kpath))) # compute all
  fo = open("KDOS_INTERFACE.OUT","w")
  fo.write("# k, E, Bulk1, Surf1, Bulk2, Surf2, interface\n")
  for o in out: fo.write(o)
  fo.close()







def surface(h1,energies=np.linspace(-1.,1.,100),operator=None,
                    delta=0.01,kpath=None,hs=None):
  """Get the surface DOS of an interface"""
  from scipy.sparse import csc_matrix,bmat
  if kpath is None: 
    if h1.dimensionality==3:
      g2d = h1.geometry.copy() # copy Hamiltonian
      g2d = sculpt.set_xy_plane(g2d)
      kpath = klist.default(g2d,nk=len(energies))
    elif h1.dimensionality==2:
      kpath = [[k,0.,0.] for k in np.linspace(0.,1.,len(energies))]
    elif h1.dimensionality==1: kpath = [[0.,0.,0.0]] # one dummy point
    else: raise
  fo = open("KDOS.OUT","w")
  fo.write("# k, E, Surface, Bulk\n")
  tr = timing.Testimator("KDOS") # generate object
  ik = 0
  h1 = h1.get_multicell() # multicell Hamiltonian
  for k in kpath:
    tr.remaining(ik,len(kpath)) # generate object
    ik += 1
    outs = green.surface_multienergy(h1,k=k,energies=energies,delta=delta,hs=hs)
    for (energy,out) in zip(energies,outs):
      # write everything
      if h1.dimensionality==1: fo.write(str(energy)+"   ")
      else: fo.write(str(ik)+"   "+str(energy)+"   ")
      for g in out: # loop
        if operator is None: d = -algebra.trace(g).imag # only the trace 
        elif callable(operator): d = operator(g,k=k) # call the operator
        else:  d = -algebra.trace(g@operator).imag # assume it is a matrix
        fo.write(str(d)+"   ") # write in a file
      fo.write("\n") # next line
      fo.flush() # flush
  fo.close()





def add(h1,h2):
    mo1 = h1.get_multihopping()
    mo2 = h2.get_multihopping()
    h3 = h1.copy() # copy
    h3.set_multihopping(mo1+mo2)
    return h3

def rmul(h,a):
    mo = h.get_multihopping()
    h1 = h.copy() # copy
    h1.set_multihopping(a*mo)
    return h1

def mul(h,a):
    mo = h.get_multihopping()
    h1 = h.copy() # copy
    h1.set_multihopping(mo*a)
    return h1




# specialized routine to perform an SCF, taking as starting point an
# attractive local interaction in a spinless Hamiltonian

from ..sctk.spinless import onsite_delta_vev
from .. import inout
import numpy as np
import time
import os

mf_file = "MF.pkl" 

def attractive_hubbard(h0,mf=None,mix=0.9,g=0.0,nk=8,solver="plain",
        maxerror=1e-5,**kwargs):
    """Perform the SCF mean field"""
    if not h0.check_mode("spinless"): raise # sanity check
    h = h0.copy() # initial Hamiltonian
    if mf is None:
      try: dold = inout.load(mf_file) # load the file
      except: dold = np.random.random(h.intra.shape[0]) # random guess
    else: dold = mf # initial guess
    ii = 0
    os.system("rm -f STOP") # remove stop file
    def f(dold):
      """Function to minimize"""
#      print("Iteration #",ii) # Iteration
      h = h0.copy() # copy Hamiltonian
      if os.path.exists("STOP"): return dold
      h.add_swave(dold*g) # add the pairing to the Hamiltonian
      t0 = time.time()
      d = onsite_delta_vev(h,nk=nk,**kwargs) # compute the pairing
      t1 = time.time()
      print("Time in this iteration = ",t1-t0) # Difference
      diff = np.max(np.abs(d-dold)) # compute the difference
      print("Error = ",diff) # Difference
      print("Average Pairing = ",np.mean(np.abs(d))) # Pairing
      print("Maximum Pairing = ",np.max(np.abs(d))) # Pairing
      print()
#      ii += 1
      return d
    if solver=="plain":
      do_scf = True
      while do_scf:
        d = f(dold) # new vector
        dold = mix*d + (1-mix)*dold # redefine
        diff = np.max(np.abs(d-dold)) # compute the difference
        if diff<maxerror: 
          do_scf = False
    else:
        print("Solver used:",solver)
        import scipy.optimize as optimize 
        if solver=="newton": fsolver = optimize.newton_krylov
        elif solver=="anderson": fsolver = optimize.anderson
        elif solver=="broyden": fsolver = optimize.broyden2
        elif solver=="linear": fsolver = optimize.linearmixing
        else: raise
        def fsol(x): return x - f(x) # function to solve
        dold = fsolver(fsol,dold,f_tol=maxerror)
    h = h0.copy() # copy Hamiltonian
    h.add_swave(dold*g) # add the pairing to the Hamiltonian
    inout.save(dold,mf_file) # save the mean field
    scf = SCF() # create object
    scf.hamiltonian = h # store
    return scf # return SCF object



class SCF(): pass




# variety of tricks to accelerate a mean field calculation
from . import hubbard


def scf_accelerate(scf):
    """Redefine methods to accelerate a mean field calculation"""
    if scf.mode=="Hubbard" and type(self.g)==type(0.2): # Hubbard mean field
        return hubbard.hubbardscf(scf.hamiltonian,U=scf.g,
                mix=scf.mixing,mf=scf.mf)
    return scf






from scipy.sparse import coo_matrix,bmat
from .rotate_spin import sx,sy,sz
from .increase_hilbert import get_spinless2full,get_spinful2full
import numpy as np
from . import checkclass
from . import geometry

def float2array(z):
    if checkclass.is_iterable(z): return z # iterable, input is an array
    else: return [0.,0.,z] # input is a number

def add_zeeman(h,zeeman=[0.0,0.0,0.0]):
  """ Add Zeeman to the hamiltonian """
  # convert the input into a list
  def evaluate_J(z,r,i):
    if checkclass.is_iterable(z): # it is a list/array
        if checkclass.is_iterable(z[0]):  # each element is a list/array
            return np.array(z[i]) # iterable, input is an array
        out = [0.,0.,0.] # not iterable
        for j in range(len(z)): # loop over elements
            if callable(z[j]):  # each element is a function
               out[j] = z[j](r) # call the function
            else: # if it is number
               out[j] = z[j]
        return np.array(out)
    elif callable(z): # it is a function
        m = z(r) # call
        if checkclass.is_iterable(m): return np.array(m) # it is an array
        else: return np.array([0.,0.,m]) # number
    else: return np.array([0.,0.,z]) # just a number
  from scipy.sparse import coo_matrix as coo
  from scipy.sparse import bmat
  if not h.has_spin:  h.turn_spinful()
  no = len(h.geometry.r) # number of orbitals (without spin)
  # create matrix to add to the hamiltonian
  bzee = [[None for i in range(no)] for j in range(no)]
  # assign diagonal terms
  r = h.geometry.r  # z position
  for i in range(no):
      JJ = evaluate_J(zeeman,r[i],i) # evaluate the exchange
      bzee[i][i] = JJ[0]*sx+JJ[1]*sy+JJ[2]*sz
  bzee = bmat(bzee) # create matrix
  h.intra = h.intra + h.spinful2full(bzee) # Add matrix 





def add_antiferromagnetism(h,m):
  """ Adds to the intracell matrix an antiferromagnetic imbalance """
  if not h.has_spin: h.turn_spinful()
  intra = h.intra # intracell hopping
  if h.geometry.has_sublattice: pass  # if has sublattice
  else: # if does not have sublattice
#    print("WARNING, no sublattice present")
    return 0. # if does not have sublattice
  if h.has_spin:
    natoms = len(h.geometry.x) # number of atoms
    out = [[None for j in range(natoms)] for i in range(natoms)] # output matrix
    # create the array
    if checkclass.is_iterable(m): # iterable, input is an array
      if len(m)!=len(h.geometry.r): raise
      mass = m # use the input array
    elif callable(m): # input is a function
      mass = [m(h.geometry.r[i]) for i in range(natoms)] # call the function
    else: # assume it is a float
      mass = [m for i in range(natoms)] # create list
    for i in range(natoms): # loop over atoms
      mi = mass[i] # select the element
      # add contribution to the Hamiltonian
      mi = float2array(mi) # convert to array
      out[i][i] = (sx*mi[0] + sy*mi[1] + sz*mi[2])*h.geometry.sublattice[i]
    out = bmat(out) # turn into a matrix
    h.intra = h.intra + h.spinful2full(out) # Add matrix 
  else:
    print("no AF for unpolarized hamiltonian")
    raise






def add_magnetism(h,m):
  """ Adds magnetism to the intracell hopping"""
  intra = h.intra # intracell hopping
  if h.has_spin:
    natoms = len(h.geometry.r) # number of atoms
    # create the array
    out = [[None for j in range(natoms)] for i in range(natoms)] # output matrix
    if checkclass.is_iterable(m):
      if checkclass.is_iterable(m[0]) and len(m)==natoms: # input is an array
        mass = m # use as arrays
      elif len(m)==3: # single exchange provided
        mass = [m for i in range(natoms)] # use as arrays
      else: raise
    elif callable(m): # input is a function
      mass = [m(h.geometry.r[i]) for i in range(natoms)] # call the function
    else: 
      print("Wrong input in add_magnetism")
      raise 
    for i in range(natoms): # loop over atoms
      mi = mass[i] # select the element
      # add contribution to the Hamiltonian
      out[i][i] = sx*mi[0] + sy*mi[1] + sz*mi[2]
    out = bmat(out) # turn into a matrix
    h.intra = h.intra + h.spinful2full(out) # Add matrix 
  else:
    print("no AF for unpolarized hamiltonian")
    raise





def add_frustrated_antiferromagnetism(h,m):
  """Add frustrated magnetism"""
  if h.geometry.sublattice_number==3:
    g = geometry.kagome_lattice()
  elif h.geometry.sublattice_number==4:
    g = geometry.pyrochlore_lattice()
    g.center()
  else: raise # not implemented
  ms = []
  for i in range(len(h.geometry.r)): # loop
    ii = h.geometry.sublattice[i] # index of the sublattice
    if callable(m):
      ms.append(-g.r[int(ii)]*m(h.geometry.r[i])) # save this one
    else:
      ms.append(-g.r[int(ii)]*m) # save this one
  h.add_magnetism(ms) # add the magnetization





def compute_magnetization(h,**kwargs):
  """Return the magnetization of the system"""
  if not h.has_spin: raise # meaningless
  if h.has_eh: raise # not implemented
  from .densitymatrix import full_dm
  dm = full_dm(h,**kwargs) # compute density matrix
  n = dm.shape[0]//2 # number of orbitals
  mz = np.array([dm[2*i,2*i] - dm[2*i+1,2*i+1] for i in range(n)])/2..real
  mx = np.array([dm[2*i,2*i+1].real for i in range(n)])
  my = np.array([dm[2*i,2*i+1].imag for i in range(n)])
  return (mx,my,mz)




# specialized routine to perform an SCF, taking as starting point an
# attractive local interaction in a spinless Hamiltonian

from .. import inout
import numpy as np
import time
import os
from .. import densitymatrix
from copy import deepcopy
from numba import jit
from .. import utilities
from ..multihopping import MultiHopping
from .. import algebra

class Interaction():
    def __init__(self,h=None):
        self.dimensionality = 0
        if h is not None: self.dimensionality = h.dimensionality
        self.v_dict = dict() # store dictionary
    def __mult__(self,a):
        """Function to multiply"""
        out = 0
        for key in self: out = out + self[key]*a[key]
        return out



def normal_term(v,dm):
    """Return the normal term of the mean field"""
    out = dm*0.0 # initialize
    return normal_term_jit(v,dm,out) # return the normal term



def normal_term_ii(v,dm):
    """Return the normal term of the mean field"""
    out = dm*0.0 # initialize
    return normal_term_ii_jit(v,dm,out) # return the normal term


def normal_term_jj(v,dm):
    """Return the normal term of the mean field"""
    out = dm*0.0 # initialize
    return normal_term_jj_jit(v,dm,out) # return the normal term


def normal_term_ij(v,dm):
    """Return the normal term of the mean field"""
    out = dm*0.0 # initialize
    return normal_term_ij_jit(v,dm,out) # return the normal term


def normal_term_ji(v,dm):
    """Return the normal term of the mean field"""
    out = dm*0.0 # initialize
    return normal_term_ji_jit(v,dm,out) # return the normal term

@jit(nopython=True)
def normal_term_jit(v,dm,out):
    """Return the normal terms, jit function"""
    n = len(v[0])
    for i in range(n): # loop
      for j in range(n): # loop
        out[i,j] = out[i,j] - v[i,j]*dm[j,i]
        out[j,i] = out[j,i] - v[i,j]*dm[i,j]
        out[i,i] = out[i,i] + v[i,j]*dm[j,j]
        out[j,j] = out[j,j] + v[i,j]*dm[i,i]
    return out


@jit(nopython=True)
def normal_term_ii_jit(v,dm,out):
    """Return the normal terms, jit function"""
    n = len(v[0])
    for i in range(n): # loop
      for j in range(n): # loop
        out[i,i] = out[i,i] + v[i,j]*dm[j,j]
    return out

@jit(nopython=True)
def normal_term_jj_jit(v,dm,out):
    """Return the normal terms, jit function"""
    n = len(v[0])
    for i in range(n): # loop
      for j in range(n): # loop
        out[j,j] = out[j,j] + v[i,j]*dm[i,i]
    return out


@jit(nopython=True)
def normal_term_ij_jit(v,dm,out):
    """Return the normal terms, jit function"""
    n = len(v[0])
    for i in range(n): # loop
      for j in range(n): # loop
        out[i,j] = out[i,j] - v[i,j]*dm[j,i]
    return out


@jit(nopython=True)
def normal_term_ji_jit(v,dm,out):
    """Return the normal terms, jit function"""
    n = len(v[0])
    for i in range(n): # loop
      for j in range(n): # loop
        out[j,i] = out[j,i] - v[i,j]*dm[i,j]
    return out



def update_hamiltonian(tdict,mf):
    """Update the hoppings with the mean field"""
    return (MultiHopping(tdict) + MultiHopping(mf)).get_dict()


def mix_mf(mf,mf0,mix=0.8):
    """Mix mean fields"""
    return ((1-mix)*MultiHopping(mf0) + mix*MultiHopping(mf)).get_dict()



def diff_mf(mf0,mf):
    """Difference mean fields"""
    out = 0.0 # initialize
    for key in mf: # loop
        if key not in mf0: out += np.mean(np.abs(mf[key]))
        else: out += np.mean(np.abs(mf0[key] - mf[key])) # add contribution
        #out += np.mean(np.abs(mf0[key] - mf[key])) # add contribution
    return out # return


def hamiltonian2dict(h):
    return h.get_dict() # return dictionary


def set_hoppings(h,hop):
    """Add the hoppings to the Hamiltonian"""
    h.set_multihopping(MultiHopping(hop))

def get_dm(h,v,nk=1):
    """Get the density matrix"""
    ds = [(0,0,0)] # directions
    if h.dimensionality>0:
        for key in v: ds.append(key) # store
    dms = h.get_density_matrix(ds=ds,nk=nk) # get all the density matrices
    return dms # return dictionary



def get_mf(v,dm,has_eh=False,compute_anomalous=True,
        compute_normal=True,**kwargs):
    """Get the mean field matrix"""
    if has_eh:
        # let us assume that it is a full Nambu spinor
        # (this may not be general, but good enough in the meantime)
        from .. import superconductivity
        dme = dict() # dictionary
        dma01 = dict() # dictionary
        dma10 = dict() # dictionary
        ns = v[(0,0,0)].shape[0]//2 # number of spinless sites
#        op = superconductivity.nambu_anomalous_reordering(ns)
#        op = op@op # comment this to go back to the previous version
        for key in dm: # extract the electron part 
#            m = op.T@dm[key]@op # transform to the new basis
            m = dm[key] # transform to the new basis
            dme[key] = superconductivity.get_eh_sector(m,i=0,j=0)
            # this is a workaround for the reordering of Nambu spinors
            dma10[key] = superconductivity.get_eh_sector(m,i=0,j=1)
        mfe = get_mf_normal(v,dme,**kwargs) # electron part of the mean field
        # anomalous part
        #dma01,dma10 = enforce_eh_symmetry_anomalous(dma01,dma10)
        mfa01 = get_mf_anomalous(v,dma10) 
        mfa01,mfa10 = enforce_eh_symmetry_anomalous(mfa01)
        ##############################################
        # now rebuild the Hamiltonian
        mf = dict()
        for key in v:
            if not compute_normal: mfe[key] = mfe[key]*0.0
            if compute_anomalous:
                m = superconductivity.build_nambu_matrix(mfe[key],
                    c12 = mfa10[key],c21=mfa01[key])
            else:
                m = superconductivity.build_nambu_matrix(mfe[key])
            mf[key] = m # store this matrix
    #        print(key)
    #        print(np.round(m,2))
            #print(np.unique(np.round(m,2)))
        if not MultiHopping(mf).is_hermitian(): # just a sanity check
            print("Non-Hermitian mean field")
            print(np.round(mf[(0,0,0)],2))
            exit()
   #     exit()
        # enforce electron-hole symmetry
   #     mf = superconductivity.enforce_eh_symmetry(mf)
        return mf # return mean field matrix
    else: return get_mf_normal(v,dm,**kwargs) # no BdG Hamiltonian




def get_mf_normal(v,dm,compute_dd=True,add_dagger=True,
        compute_cross=True):
    """Get the mean field"""
    zero = dm[(0,0,0)]*0. # zero
    mf = dict()
    for d in v: mf[d] = zero.copy()  # initialize
    # compute the contribution to the mean field
    # onsite term
#    mf[(0,0,0)] = normal_term(v[(0,0,0)],dm[(0,0,0)]) 
    def dag(m): return m.T.conjugate()
    for d in v: # loop over directions
        d2 = (-d[0],-d[1],-d[2]) # minus this direction
        # add the normal terms
        if compute_cross: # only density density terms
            m = normal_term_ij(v[d],dm[d2]) # get matrix
            mf[d] = mf[d] + m # add normal term
            if add_dagger:
                mf[d2] = mf[d2] + dag(m) # add normal term
        if compute_dd: # density density terms
            m = normal_term_ii(v[d],dm[(0,0,0)]) # get matrix
            mf[(0,0,0)] = mf[(0,0,0)] + m # add normal term
            m = normal_term_jj(v[d2],dm[(0,0,0)]) # get matrix
            mf[(0,0,0)] = mf[(0,0,0)] + m # add normal term
    return mf



# anomalous part of the mean field
from .superscf import get_mf_anomalous
from .superscf import enforce_eh_symmetry_anomalous






def get_dc_energy(v,dm):
    """Compute double counting energy"""
    out = 0.0
    for d in v: # loop over interactions
        d2 = (-d[0],-d[1],-d[2]) # minus this direction
        n = v[d].shape[0] # shape
        for i in range(n): # loop
          for j in range(n): # loop
              out -= v[d][i,j]*dm[(0,0,0)][i,i]*dm[(0,0,0)][j,j]
              c = dm[d][i,j] # cross term
              out += v[d][i,j]*c*np.conjugate(c) # add contribution
#    print("DC energy",out.real)
    return out.real



from .mfconstrains import obj2mf

mf_file = "MF.pkl" 

def generic_densitydensity(h0,mf=None,mix=0.1,v=None,nk=8,solver="plain",
        maxerror=1e-5,filling=None,callback_mf=None,callback_dm=None,
        load_mf=True,compute_cross=True,compute_dd=True,verbose=1,
        compute_anomalous=True,compute_normal=True,info=False,
        callback_h=None,**kwargs):
    """Perform the SCF mean field"""
    if verbose>1: info=True
#    if not h0.check_mode("spinless"): raise # sanity check
    h1 = h0.copy() # initial Hamiltonian
    h1.turn_dense()
    h1.nk = nk # store the number of kpoints
    if mf is None: # no mean field given
      try: 
          if load_mf: 
              mf = inout.load(mf_file) # load the file
              MultiHopping(h0.get_dict()) + MultiHopping(mf) # see if compatible
          else: raise
      except: 
          mf = dict()
          for d in v: mf[d] = np.exp(1j*np.random.random(h1.intra.shape))
          mf[(0,0,0)] = mf[(0,0,0)] + mf[(0,0,0)].T.conjugate()
    elif type(mf)==str:
        from ..meanfield import guess
        mf = guess(h0,mode=mf) # overwrite
    else: pass # assume that it is a valid mf
    mf = obj2mf(mf) # convert to MF
    ii = 0
    os.system("rm -f STOP") # remove stop file
    hop0 = hamiltonian2dict(h1) # create dictionary
    def f(mf,h=h1):
      """Function to minimize"""
#      print("Iteration #",ii) # Iteration
      mf0 = deepcopy(mf) # copy
      h = h1.copy()
      hop = update_hamiltonian(hop0,mf) # add the mean field to the Hamiltonian
      set_hoppings(h,hop) # set the new hoppings in the Hamiltonian
      if callback_h is not None:
          h = callback_h(h) # callback for the Hamiltonian
      t0 = time.perf_counter() # time
      dm = get_dm(h,v,nk=nk) # get the density matrix
      if callback_dm is not None:
          dm = callback_dm(dm) # callback for the density matrix
      t1 = time.perf_counter() # time
      # return the mean field
      mf = get_mf(v,dm,compute_cross=compute_cross,compute_dd=compute_dd,
              has_eh=h0.has_eh,compute_anomalous=compute_anomalous,
              compute_normal=compute_normal) 
      if callback_mf is not None:
          mf = callback_mf(mf) # callback for the mean field
      t2 = time.perf_counter() # time
      if verbose>1: print("Time in density matrix = ",t1-t0) # Difference
      if verbose>1: print("Time in the normal term = ",t2-t1) # Difference
      scf = SCF() # create object
      scf.hamiltonian = h # store
#      h.check() # check the Hamiltonian
      scf.hamiltonian0 = h0 # store
      scf.mf = mf # store mean field
      if os.path.exists("STOP"): scf.mf = mf0 # use the guess
      scf.dm = dm # store density matrix
      scf.v = v # store interaction
      scf.tol = maxerror # maximum error
      return scf
    if solver=="plain":
      do_scf = True
#      from .mixing import Mixing
#      Mxg = Mixing() # initialize
      while do_scf:
        scf = f(mf) # new vector
        mfnew = scf.mf # new vector
        t0 = time.perf_counter() # time
        diff = diff_mf(mfnew,mf) # mix mean field
#        mix = Mxg.get_mix(diff) # add error
#        print("Mixing",mix)
        mf = mix_mf(mfnew,mf,mix=mix) # mix mean field
        if callback_mf is not None: # redefine mean-field if necessary
            mf = callback_mf(mf) # callback for the mean field
        t1 = time.perf_counter() # time
        if verbose>1: print("Time in mixing",t1-t0)
        if verbose>0: print("ERROR in the SCF cycle",diff)
        #print("Mixing",dmix)
        if diff<maxerror: 
            scf = f(mfnew) # last iteration, with the unmixed mean field
            inout.save(scf.mf,mf_file) # save the mean field
         #   scf.hamiltonian.check(tol=100*maxerror) # perform some sanity checks
            return scf
    else: # use different solvers
        scf = f(mf) # perform one iteration
        fmf2a = get_mf2array(scf) # convert MF to array
        fa2mf = get_array2mf(scf) # convert array to MF
        def fsol(x): # define the function to solve
            mf1 = fa2mf(x) # convert to a MF
            scf1 = f(mf1) # compute function
            xn = fmf2a(scf1.mf) # new vector
            diff = x - xn # difference vector
            print("ERROR",np.max(np.abs(diff)))
            print()
            return x - xn # return vector
        x0 = fmf2a(scf.mf) # initial guess
        # these methods do seem too efficient, but lets have them anyway
        if solver=="krylov":
            from scipy.optimize import newton_krylov
            x = newton_krylov(fsol,x0,rdiff=1e-3) # use the solver
        elif solver=="anderson":
            from scipy.optimize import anderson
            x = anderson(fsol,x0) # use the solver
        elif solver=="broyden1":
            from scipy.optimize import broyden1
            x = broyden1(fsol,x0,f_tol=maxerror*100) # use the solver
        elif solver=="linear":
            from scipy.optimize import linearmixing
            x = linearmixing(fsol,x0,f_tol=maxerror*100) # use the solver
        else: raise # unrecognised solver
        mf = fa2mf(x) # transform to MF
        scf = f(mf) # compute the SCF with the solution
        scf.error = maxerror # store the error
        inout.save(scf.mf,mf_file) # save the mean field
        return scf # return the mean field


def get_mf2array(scf):
    """Function to transform the mean field in an array"""
    nt = len(scf.mf) # number of terms in the dictionary
    n = scf.mf[(0,0,0)].shape[0]
    def fmf2a(mf):
        #print(mf[(0,0,0)].real)
        out = [mf[key].real for key in mf] # to plain array
        out += [mf[key].imag for key in mf] # to plain array
        out = np.array(out)
#        print(out.shape)
        out = out.reshape(nt*n*n*2) # reshape
        return out
    return fmf2a # return function

def get_array2mf(scf):
    """Function to transform an array into a mean field"""
    ds = [key for key in scf.mf] # store keys
    nt = len(scf.mf) # number of terms in the dictionary
    n = scf.mf[(0,0,0)].shape[0] # size
    def fa2mf(a):
        a = a.copy().reshape((2*nt,n*n)) # reshape array
        mf =  dict()
        #print(a.shape)
        for i in range(len(ds)):
            d = ds[i]
            m = a[i,:] + 1j*a[i+nt,:] # get matrix
            mf[d] = m.reshape((n,n)) # store
        return mf
    return fa2mf # return function


def densitydensity(h,filling=0.5,mu=None,verbose=0,**kwargs):
    """Function for density-density interactions"""
    if h.has_eh: 
        if not h.has_spin: return NotImplemented # only for spinful
    h = h.get_multicell()
    h.turn_dense()
    def callback_h(h):
        """Set the filling"""
        if mu is None:
          fermi = h.get_fermi4filling(filling,nk=h.nk) # get the filling
          if verbose>1: print("Fermi energy",fermi)
          h.fermi = fermi
          h.shift_fermi(-fermi) # shift by the fermi energy
        else: h.shift_fermi(-mu) # shift by mu
        return h
#    callback_h = None
    scf = generic_densitydensity(h,callback_h=callback_h,verbose=verbose,
            **kwargs)
    # Now compute the total energy
    h = scf.hamiltonian
    etot = h.get_total_energy(nk=h.nk)
    if mu is None: 
        etot += h.fermi*h.intra.shape[0]*filling # add the Fermi energy
    #print("Occupied energies",etot)
    etot += get_dc_energy(scf.v,scf.dm) # add the double counting energy
    etot = etot.real
    scf.total_energy = etot
    if verbose>1:
      print("##################")
      print("Total energy",etot)
      print("##################")
    return scf




def hubbard(h,U=1.0,constrains=[],**kwargs):
    """Wrapper to perform a Hubbard model calculation"""
    h = h.copy() # copy Hamiltonian
    h.turn_multicell() # multicell Hamiltonian
    U = obj2geometryarray(U,h.geometry) # redefine as array 
    n = len(h.geometry.r) # number of spinless sites
    if h.has_spin:
      zero = np.zeros((2*n,2*n),dtype=np.complex)
      for i in range(n): zero[2*i,2*i+1] = U[i] # Hubbard interaction
    else: 
      zero = np.zeros((n,n),dtype=np.complex)
      n = len(h.geometry.r) # number of spinless sites
      for i in range(n): zero[i,i] = U[i] # Hubbard interaction
    v = dict() # dictionary
    v[(0,0,0)] = zero 
    from . import mfconstrains
    def callback_mf(mf):
        """Put the constrains in the mean field if necessary"""
        mf = mfconstrains.enforce_constrains(mf,h,constrains)
        return mf
    if h.has_spin:
      return densitydensity(h,v=v,callback_mf=callback_mf,**kwargs)
    else:
      return densitydensity(h,v=v,compute_cross=False,
              callback_mf=callback_mf,**kwargs)


def Vinteraction(h,V1=0.0,V2=0.0,V3=0.0,U=0.0,
        constrains=[],Vr=None,**kwargs):
    """Perform a mean-field calculation with density-density interactions
    - U, local Hubbard interaction
    - V1, first neighbor interaction
    - V2, second neighbor interaction
    """
    h = h.get_multicell() # multicell Hamiltonian
    h.turn_dense()
    # define the function
    nd = h.geometry.neighbor_distances() # distance to first neighbors
    from .. import specialhopping
    mgenerator = specialhopping.distance_hopping_matrix([V1/2.,V2/2.,V3/2.],nd[0:3])
    hv = h.geometry.get_hamiltonian(has_spin=False,is_multicell=True,
            mgenerator=mgenerator) 
    if Vr is not None:
      hv1 = h.geometry.get_hamiltonian(has_spin=False,is_multicell=True,
              tij=Vr)
      hv = hv + hv1 # add the two Hamiltonians
    v = hv.get_hopping_dict() # hopping dictionary
    U = obj2geometryarray(U,h.geometry) # convert to array
    if h.has_spin: #raise # not implemented
        for d in v: # loop
            m = v[d] ; n = m.shape[0]
            m1 = np.zeros((2*n,2*n),dtype=np.complex)
            for i in range(n):
              for j in range(n): 
                  m1[2*i,2*j] = m[i,j]
                  m1[2*i+1,2*j] = m[i,j]
                  m1[2*i,2*j+1] = m[i,j]
                  m1[2*i+1,2*j+1] = m[i,j]
            v[d] = m1 # store
        for i in range(n):
            v[(0,0,0)][2*i,2*i+1] += U[i]/2. # add
            v[(0,0,0)][2*i+1,2*i] += U[i]/2. # add
    # Now put the constrains if necessary
    from . import mfconstrains
    def callback_mf(mf):
        """Put the constrains in the mean field if necessary"""
        mf = mfconstrains.enforce_constrains(mf,h,constrains)
        return mf
    return densitydensity(h,v=v,callback_mf=callback_mf,**kwargs)



from ..meanfield import identify_symmetry_breaking

class SCF():
    def identify_symmetry_breaking(self,**kwargs):
        return identify_symmetry_breaking(self.hamiltonian,self.hamiltonian0,
                tol=10*self.tol,**kwargs)
    def order_parameter(self,name):
        from ..meanfield import order_parameter
        return order_parameter(self,name)





def obj2geometryarray(U,g):
    """Convert an object to an array"""
    if algebra.isnumber(U):
        return np.array([U for ir in g.r]) # same for all
    elif callable(U):
        return np.array([U(ir) for ir in g.r]) # call for each
    else: raise



from . import ribbonizate
from . import multicell
from . import hamiltonians
import numpy as np


def bar_generator(h,n=50,sparse=True):
  """Return a generator of quantum Hall bars, input is magnetic field"""
# create hamiltonian
  hr = multicell.bulk2ribbon(h,n=n,sparse=sparse,nxt=6) 
  def fun(b):
    """Return Hamiltonian of the bar"""
    ho = hr.copy()
    ho.add_peierls(b) # add magnetic field
    return ho
  return fun # return function







def bulk2ribbon(h,mag_field=0.0,n=10,sparse=True,check=True):
  """ Generate the Hamiltonian of a ribbon with magnetic field"""
  if not h.is_multicell: # dense hamiltonian
    ho = ribbonizate.hamiltonian_bulk2ribbon(h,n=n,sparse=sparse,check=check) # create hamiltonian
  else: # sparse hamiltonian
    ho = multicell.bulk2ribbon(h,n=n,sparse=sparse,nxt=6) # create hamiltonian
  ho.geometry.center()
  if np.abs(mag_field)>0.0000001: ho.add_peierls(mag_field) # add peierls phase
  return ho


def landau_levels(h,mag_field=0.01,k=0.0,nl=10,rfactor=1.5):
  """Return the energies of the Landau levels for a certain magnetic
  field"""
  nc = int(round(rfactor*h.geometry.a2[1]/mag_field)) # number of replicas
  hr = bulk2ribbon(h,mag_field=mag_field,n=nc,sparse=True) # generate ribbon
  from . import operators
  # now get the 
  operator = operators.bulk1d(hin,p=0.7)
  hkgen = hin.get_hk_gen()
  hk = hkgen(k) # get hamiltonian
  # eigenvalues and eigenvectors
  eig,eigvec = lg.eigsh(hk,k=nbands,which="LM",sigma=0.0) 
  eigvec = eigvec.transpose() # tranpose the matrix
  eout = [] # output energies
  for (e,v) in zip(eig,eigvec): # loop over eigenvectors
    v = csc_matrix(v)
    a = np.conjugate(v) * operator * v.T
    a = a.data[0]
    a = a.real # real part
    if np.abs(a)<0.1: eout.append(e)
  return eout # return energies of landau levels



import numpy as np

def multireplicas(self,n):
    """
    Return several replicas of the unit cell, it is similar
    to supercell but without the shift of the center and going to positive
    and negative positions
    """
    if n==0: return self.r # return positions
    out = [] # empty list with positions
    dl = self.neighbor_directions(n) # list with neighboring cells to take
    if self.dimensionality==0: return self.r
    else: # bigger dimensionality
        for d in dl:  out += self.replicas(d=d) # add this direction
    return np.array(out)

def replicas(self,d=[0.,0.,0.]):
    """Return replicas of the atoms in the unit cell"""
    return np.array([ri + self.a1*d[0] + self.a2*d[1] + self.a3*d[2] for ri in self.r])



import numpy as np
from . import klist
from . import algebra
import scipy.linalg as lg


def ipr2d(h,nk=10,random=True,window=[-0.1,0.1]):
  """Calculate the inverse participation ratio"""
  hkgen = h.get_hk_gen() # get generator
  ks = range(nk*nk)
  out = 0.0 # start
  for k in ks:
    k = np.random.random(2) # random k-point
    hk = hkgen(k) # get hamiltonian
    es,ws = lg.eigh(hk) # diagonalize
    ws = ws.transpose() # transpose
    nume = 0
    tmp = 0.0
    for (e,w) in zip(es,ws):
      if window[0]<e<window[1]:
        nume += 1
        d = w*np.conjugate(w) # density
        d2 = (d*d).real # ipr
        d2 = np.sum(d2) # ipr
        tmp += d2 # store
    if nume>0: tmp/= nume # normalize
    out += tmp # add
  return out/len(ks)
        





def ipr(m,retain=lambda x: True):
  """Calculate the inverse participation ratio"""
  es,ws = algebra.eigh(m) # diagonalize
  ws = ws.transpose() # transpose
  eout = []
  dout = []
  for (ie,iw) in zip(es,ws): # loop
    if retain(iw): # if retain
        dout.append(np.sum(np.abs(iw)**4))
        eout.append(ie) # store energy
  return np.array(eout),np.array(dout) # return value



import numpy.linalg as lg
from scipy.optimize import minimize_scalar
import scipy.optimize as opt
import numpy as np
import scipy.sparse.linalg as lgs
from scipy.sparse import csc_matrix
from . import algebra

def minimize_gap(f,tol=0.001,bounds=(0,1.)):
  """Minimizes the gap of the system, the argument is between 0 and 1"""
  return f(minimize_scalar(f,method="Bounded",bounds=bounds,tol=tol).x)





def gap_line(h,kpgen,assume_eh = False,sparse=True,nocc=None):
  """Return a function with argument between 0,1, which returns the gap"""
  hk_gen = h.get_hk_gen() # get hamiltonian generator
  def f(k):
    kp = kpgen(k) # get kpoint
    hk = hk_gen(kp) # generate hamiltonian
    if sparse: 
      es,ew = lgs.eigsh(csc_matrix(hk),k=4,which="LM",sigma=0.0)
    else:
      es = lg.eigvalsh(hk) # get eigenvalues
    if assume_eh: g = np.min(es[es>0.])
    else:  
      if nocc is None: # Assume conduction are staets above E = 0.0
        try: g = np.min(es[es>0.]) - np.max(es[es<0.])
        except: g = 0.0 
      else:
        g = es[nocc] - es[nocc-1]
    return g  # return gap
  return f  # return gap


def raw_gap(h,kpgen,sparse=True,nk=100):
  hk_gen = h.get_hk_gen() # get hamiltonian generator
  ks = np.linspace(0.,1.,nk)
  etot = [] # total eigenvalues
  for k in ks:
    kp = kpgen(k)
    hk = hk_gen(kp) # generate hamiltonian
    if sparse: 
      es,ew = lgs.eigsh(csc_matrix(hk),k=4,which="LM",sigma=0.0)
    else:
      es = lg.eigvalsh(hk) # get eigenvalues
    etot.append(es)
  etot = np.array(etot)
  return min(etot[etot>0.])


def gap2d(h,nk=40,k0=None,rmap=1.0,recursive=False,
           iterations=10,sparse=True,mode="refine"):
  """Calculates the gap for a 2d Hamiltonian by doing
  a kmesh sampling. It will return the positive energy with smaller value"""
  if mode=="optimize": # using optimize library
    from scipy.optimize import minimize
    hk_gen = h.get_hk_gen() # generator
    def minfun(k): # function to minimize
      hk = hk_gen(k) # Hamiltonian 
      if h.is_sparse: 
          es = algebra.smalleig(hk,numw=10)
      else: es = algebra.eigvalsh(hk) # get eigenvalues
      ggg = np.min(es[es>0.])+np.min(np.abs(es[es<0.])) # gap
      return ggg # retain positive
    gaps = [minimize(minfun,np.random.random(h.dimensionality),method="Powell").fun  for i in range(iterations)]
#    print(gaps)
    return np.min(gaps)

  else: # classical way
    if k0 is None: k0 = np.random.random(2) # random shift
    if h.dimensionality != 2: raise
    hk_gen = h.get_hk_gen() # get hamiltonian generator
    emin = 1000. # initial values
    for ix in np.linspace(-.5,.5,nk):  
      for iy in np.linspace(-.5,.5,nk):  
        k = np.array([ix,iy]) # generate kvector
        if recursive: k = k0 + k*rmap # scale vector
        hk = hk_gen(k) # generate hamiltonian
        if h.is_sparse: 
            es = algebra.smalleig(hk,numw=4)
        else: 
            es = algebra.eigvalsh(hk) # get eigenvalues
        ggg = np.min(es[es>0.])+np.min(np.abs(es[es<0.])) # gap
        if ggg<emin:
          emin = ggg # store new minimum 
          kbest = k.copy() # store the best k
    if recursive: # if it has been chosen recursive
      if iterations>0: # if still iterations left
        emin = gap2d(h,nk=nk,k0=kbest,rmap=rmap/4,recursive=recursive,
                        iterations=iterations-1,sparse=sparse)
    return emin # gap



def optimize_gap_single(h,direct=True):
  """Return the gap, just one time"""
  hkgen = h.get_hk_gen() # get generator
  dim = h.dimensionality # dimensionality
  if direct: # returnt the direct gap
    def fg(k): # minimize the gap
      es = lg.eigvalsh(hkgen(k)) # eigenvalues
      return np.min(es[es>0.])-np.max(es[es<0.]) # return gap
    x0 = np.random.random(dim) # random point
    bounds = [(0,1.) for i in range(dim)] # bounds
    result = opt.minimize(fg,x0,bounds=bounds,method="SLSQP")
    x = result.x # position of the minimum gap
    return (fg(x),x)
  else: # indirect gap
    def fg(k): # minimize the gap
      k1 = np.array([k[2*i] for i in range(dim)])
      k2 = np.array([k[2*i+1] for i in range(dim)])
      es1 = algebra.eigvalsh(hkgen(k1)) # eigenvalues
      es2 = algebra.eigvalsh(hkgen(k2)) # eigenvalues
      return np.min(es1[es1>0.])-np.max(es2[es2<0.]) # return gap
    x0 = np.random.random(dim*2) # random point
    bounds = [(0,1.) for i in range(2*dim)] # bounds
    result = opt.minimize(fg,x0,bounds=bounds,method="SLSQP")
    x = result.x # position of the minimum gap
    return (fg(x),x)
     
def optimize_gap(h,direct=True,ntries=10):
  """Return the gap, several times"""
  rs = [optimize_gap_single(h,direct=direct) for i in range(ntries)]
  gaps = [r[0] for r in rs] # gaps
  mg = np.min(gaps) # minimum gap
  for r in rs: # loop over gaps
    if r[0]==mg: return r # return minimum



def indirect_gap(h,robust=True,mode="full",**kwargs):
    """Calculates the gap for a 2d Hamiltonian by doing
    a kmesh sampling. It will return the positive energy with smallest value"""
    if h.intra.shape[0]>2000:
        h = h.copy()
        h.turn_sparse() # sparse Hamiltonian
    from scipy.optimize import minimize
    hk_gen = h.get_hk_gen() # generator
    def gete(k): # return the energies
      hk = hk_gen(k) # Hamiltonian 
      if h.is_sparse: es = algebra.smalleig(hk,numw=3) # sparse
      else: es = algebra.eigvalsh(hk) # get eigenvalues
      return es # get the energies
    # We will assume that the chemical potential is at zero
    def func(k): # conduction band eigenvalues
      es = gete(k) # get eigenvalues
      try:
          es = es[es>0.] # conduction band
          return np.min(es) # minimum energy
      except: return 0.0
    def funv(k): # valence band eigenvalues
      es = gete(k) # get eigenvalues
      try:
          es = -es[es<0.] # valence band
          return np.min(es) # maximum energy
      except: return 0.0
    def funcv(k): # valence band eigenvalues
      es = gete(k) # get eigenvalues
      ec = np.min(es[es>0.0]) # conduction band
      ev = np.min(-es[es<0.0]) # valence band
      return ec+ev # energy difference
    def opte(f):
      """Optimize the eigenvalues"""
      from scipy.optimize import differential_evolution
      from scipy.optimize import minimize
      bounds = [(0.,1.) for i in range(h.dimensionality)]
      if robust: res = differential_evolution(f,bounds=bounds,**kwargs)
      else: 
          x0 = np.random.random(h.dimensionality) # inital vector
          res = minimize(f,x0,method="Powell",bounds=bounds,**kwargs)
      return f(res.x)
    if mode=="full":
        ev = opte(funv) # optimize valence band
        if h.has_eh: ec = ev # workaround for SC
        else: ec = opte(func) # optimize conduction band
        return ec+ev # return result
    elif mode=="valence":
        return opte(funv) # optimize valence band
    elif mode=="conduction":
        return opte(func) # optimize conduction band
    else: raise





def get_gap(self,ntries=1,**kwargs):
    """Returns the gap of the Hamiltonian"""
    return np.min([indirect_gap(self,**kwargs) for i in range(ntries)]) 

